# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

import os
import subprocess

import pytest

import isctest
from isctest.hypothesis.strategies import dns_names

from hypothesis import strategies, given, settings

pytest.importorskip("dns.dnssectypes")
from dns.dnssectypes import NSEC3Hash
import dns.dnssec

NSEC3HASH = os.environ.get("NSEC3HASH")


# test cases from RFC 5155, Appendix A
@pytest.mark.parametrize(
    "domain,nsec3hash",
    [
        ("*.w.example", "R53BQ7CC2UVMUBFU5OCMM6PERS9TK9EN"),
        (
            "2t7b4g4vsa5smi47k61mv5bv1a22bojr.example",
            "KOHAR7MBB8DC2CE8A9QVL8HON4K53UHI",
        ),
        ("a.example", "35MTHGPGCU1QG68FAB165KLNSNK3DPVL"),
        ("ai.example", "GJEQE526PLBF1G8MKLP59ENFD789NJGI"),
        ("example", "0P9MHAVEQVM6T7VBL5LOP2U3T2RP3TOM"),
        ("ns1.example", "2T7B4G4VSA5SMI47K61MV5BV1A22BOJR"),
        ("ns2.example", "Q04JKCEVQVMU85R014C7DKBA38O0JI5R"),
        ("w.example", "K8UDEMVP1J2F7EG6JEBPS17VP3N8I58H"),
        ("x.w.example", "B4UM86EGHHDS6NEA196SMVMLO4ORS995"),
        ("x.y.w.example", "2VPTU5TIMAMQTTGL4LUU9KG21E0AOR3S"),
        ("xx.example", "T644EBQK9BIBCNA874GIVR6JOJ62MLHV"),
        ("y.w.example", "JI6NEOAEPV8B5O6K4EV33ABHA8HT9FGC"),
    ],
)
def test_nsec3_hashes(domain, nsec3hash):
    salt = "aabbccdd"
    algorithm = "1"
    iterations = "12"

    cmd = isctest.run.cmd([NSEC3HASH, salt, algorithm, iterations, domain])
    assert nsec3hash in cmd.out

    flags = "0"
    cmd = isctest.run.cmd([NSEC3HASH, "-r", algorithm, flags, iterations, salt, domain])
    assert nsec3hash in cmd.out


@pytest.mark.parametrize(
    "salt_emptiness_args",
    [
        [""],
        ["-"],
        ["--", ""],
        ["--", "-"],
    ],
)
def test_nsec3_empty_salt(salt_emptiness_args):
    algorithm = "1"
    iterations = "0"
    domain = "com"

    cmd = isctest.run.cmd(
        [NSEC3HASH] + salt_emptiness_args + [algorithm, iterations, domain]
    )
    assert "CK0POJMG874LJREF7EFN8430QVIT8BSM" in cmd.out
    assert "salt=-" in cmd.out


@pytest.mark.parametrize(
    "salt_emptiness_arg",
    [
        "",
        "-",
    ],
)
def test_nsec3_empty_salt_r(salt_emptiness_arg):
    algorithm = "1"
    flags = "1"
    iterations = "0"
    domain = "com"

    cmd = isctest.run.cmd(
        [
            NSEC3HASH,
            "-r",
            algorithm,
            flags,
            iterations,
            salt_emptiness_arg,
            domain,
        ]
    )
    assert " - CK0POJMG874LJREF7EFN8430QVIT8BSM" in cmd.out


@pytest.mark.parametrize(
    "args",
    [
        [""],  # missing arg
        ["two", "names"],  # extra arg
    ],
)
def test_nsec3_missing_args(args):
    with pytest.raises(subprocess.CalledProcessError):
        isctest.run.cmd([NSEC3HASH, "00", "1", "0"] + args)


def test_nsec3_bad_option():
    with pytest.raises(subprocess.CalledProcessError):
        isctest.run.cmd([NSEC3HASH, "-?"])


@given(
    domain=dns_names(),
    it=strategies.integers(min_value=0, max_value=65535),
    salt_bytes=strategies.binary(min_size=0, max_size=255),
)
def test_nsec3hash_acceptable_values(domain, it, salt_bytes) -> None:
    if not salt_bytes:
        salt_text = "-"
    else:
        salt_text = salt_bytes.hex()
    # calculate the hash using dnspython:
    hash1 = dns.dnssec.nsec3_hash(
        domain, salt=salt_bytes, iterations=it, algorithm=NSEC3Hash.SHA1
    )

    # calculate the hash using nsec3hash:
    cmd = isctest.run.cmd([NSEC3HASH, salt_text, "1", str(it), str(domain)])
    hash2 = cmd.out.partition(" ")[0]

    assert hash1 == hash2


@settings(max_examples=5)
@given(
    domain=dns_names(),
    it=strategies.integers(min_value=0, max_value=65535),
    salt_bytes=strategies.binary(min_size=256),
)
def test_nsec3hash_salt_too_long(domain, it, salt_bytes) -> None:
    salt_text = salt_bytes.hex()
    with pytest.raises(subprocess.CalledProcessError):
        isctest.run.cmd([NSEC3HASH, salt_text, "1", str(it), str(domain)])


@settings(max_examples=5)
@given(
    domain=dns_names(),
    it=strategies.integers(min_value=65536),
    salt_bytes=strategies.binary(min_size=0, max_size=255),
)
def test_nsec3hash_too_many_iterations(domain, it, salt_bytes) -> None:
    salt_text = salt_bytes.hex()
    with pytest.raises(subprocess.CalledProcessError):
        isctest.run.cmd([NSEC3HASH, salt_text, "1", str(it), str(domain)])



# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

import shutil
from typing import List

import isctest
from isctest.kasp import private_type_record
from isctest.template import Nameserver, TrustAnchor, Zone
from isctest.run import EnvCmd
from rollover.common import default_algorithm


def configure_tld(zonename: str, delegations: List[Zone]) -> Zone:
    templates = isctest.template.TemplateEngine(".")
    alg = default_algorithm()
    keygen = EnvCmd("KEYGEN", f"-q -a {alg.number} -b {alg.bits} -L 3600")
    signer = EnvCmd("SIGNER", "-S -g")

    isctest.log.info(f"create {zonename} zone with delegations and sign")

    for zone in delegations:
        try:
            shutil.copy(f"{zone.ns.name}/dsset-{zone.name}.", "ns2/")
        except FileNotFoundError:
            # Some delegations are unsigned.
            pass

    ksk_name = keygen(f"-f KSK {zonename}", cwd="ns2").out.strip()
    zsk_name = keygen(f"{zonename}", cwd="ns2").out.strip()
    ksk = isctest.kasp.Key(ksk_name, keydir="ns2")
    zsk = isctest.kasp.Key(zsk_name, keydir="ns2")
    dnskeys = [ksk.dnskey, zsk.dnskey]

    template = "template.db.j2.manual"
    outfile = f"{zonename}.db"
    tdata = {
        "fqdn": f"{zonename}.",
        "delegations": delegations,
        "dnskeys": dnskeys,
    }
    templates.render(f"ns2/{outfile}", tdata, template=f"ns2/{template}")
    signer(f"-P -x -O full -o {zonename} -f {outfile}.signed {outfile}", cwd="ns2")

    return Zone(zonename, f"{outfile}.signed", Nameserver("ns2", "10.53.0.2"))


def configure_root(delegations: List[Zone]) -> TrustAnchor:
    templates = isctest.template.TemplateEngine(".")
    alg = default_algorithm()
    keygen = EnvCmd("KEYGEN", f"-q -a {alg.number} -b {alg.bits} -L 3600")
    signer = EnvCmd("SIGNER", "-S -g")

    zonename = "."
    isctest.log.info("create root zone with delegations and sign")

    for zone in delegations:
        shutil.copy(f"{zone.ns.name}/dsset-{zone.name}.", "ns1/")

    ksk_name = keygen(f"-f KSK {zonename}", cwd="ns1").out.strip()
    zsk_name = keygen(f"{zonename}", cwd="ns1").out.strip()
    ksk = isctest.kasp.Key(ksk_name, keydir="ns1")
    zsk = isctest.kasp.Key(zsk_name, keydir="ns1")
    dnskeys = [ksk.dnskey, zsk.dnskey]

    template = "root.db.j2.manual"
    infile = "root.db.in"
    outfile = "root.db.signed"
    tdata = {
        "fdqn": f"{zonename}.",
        "delegations": delegations,
        "dnskeys": dnskeys,
    }
    templates.render(f"ns1/{infile}", tdata, template=f"ns1/{template}")
    signer(f"-P -x -O full -o {zonename} -f {outfile} {infile}", cwd="ns1")

    return ksk.into_ta("static-ds")


def fake_lifetime(key: str, lifetime: int):
    """
    Fake lifetime of key.
    """
    with open(f"ns3/{key}.state", "a", encoding="utf-8") as statefile:
        statefile.write(f"Lifetime: {lifetime}\n")


def set_key_relationship(key1: str, key2: str):
    """
    Set in the key state files the Predecessor/Successor fields.
    """
    predecessor = isctest.kasp.Key(key1, keydir="ns3")
    successor = isctest.kasp.Key(key2, keydir="ns3")

    with open(f"ns3/{key1}.state", "a", encoding="utf-8") as statefile:
        statefile.write(f"Successor: {successor.tag}\n")

    with open(f"ns3/{key2}.state", "a", encoding="utf-8") as statefile:
        statefile.write(f"Predecessor: {predecessor.tag}\n")


def render_and_sign_zone(
    zonename: str, keys: List[str], signing: bool = True, extra_options: str = ""
):
    dnskeys = []
    privaterrs = []
    for key_name in keys:
        key = isctest.kasp.Key(key_name, keydir="ns3")
        privaterr = private_type_record(zonename, key)
        dnskeys.append(key.dnskey)
        privaterrs.append(privaterr)

    outfile = f"{zonename}.db"
    templates = isctest.template.TemplateEngine(".")
    template = "template.db.j2.manual"
    tdata = {
        "fqdn": f"{zonename}.",
        "dnskeys": dnskeys,
        "privaterrs": privaterrs,
    }
    templates.render(f"ns3/{outfile}", tdata, template=f"ns3/{template}")

    if signing:
        signer = EnvCmd("SIGNER", "-S -g -x -s now-1h -e now+2w -O raw")
        signer(
            f"{extra_options} -o {zonename} -f {outfile}.signed {outfile}", cwd="ns3"
        )


def configure_algo_csk(tld: str, policy: str, reconfig: bool = False) -> List[Zone]:
    # The zones at csk-algorithm-roll.$tld represent the various steps
    # of a CSK algorithm rollover.
    zones = []
    zone = f"csk-algorithm-roll.{tld}"
    keygen = EnvCmd("KEYGEN", f"-k {policy}")
    settime = EnvCmd("SETTIME", "-s")

    # Step 1:
    # Introduce the first key. This will immediately be active.
    zonename = f"step1.{zone}"
    zones.append(Zone(zonename, f"{zonename}.db", Nameserver("ns3", "10.53.0.3")))
    isctest.log.info(f"setup {zonename}")
    TactN = "now-7d"
    TsbmN = "now-161h"
    csktimes = f"-P {TactN} -A {TactN}"
    # Key generation.
    csk_name = keygen(f"-l csk1.conf {csktimes} {zonename}", cwd="ns3").out.strip()
    settime(
        f"-g OMNIPRESENT -k OMNIPRESENT {TactN} -r OMNIPRESENT {TactN} -z OMNIPRESENT {TactN} -d OMNIPRESENT {TactN} {csk_name}",
        cwd="ns3",
    )
    # Signing.
    render_and_sign_zone(zonename, [csk_name], extra_options="-z")

    if reconfig:
        # Step 2:
        # After the publication interval has passed the DNSKEY is OMNIPRESENT.
        zonename = f"step2.{zone}"
        zones.append(Zone(zonename, f"{zonename}.db", Nameserver("ns3", "10.53.0.3")))
        isctest.log.info(f"setup {zonename}")
        # The time passed since the new algorithm keys have been introduced is 3 hours.
        TpubN1 = "now-3h"
        csktimes = f"-P {TactN} -A {TactN} -P sync {TsbmN} -I now"
        newtimes = f"-P {TpubN1} -A {TpubN1}"
        # Key generation.
        csk1_name = keygen(f"-l csk1.conf {csktimes} {zonename}", cwd="ns3").out.strip()
        csk2_name = keygen(f"-l csk2.conf {newtimes} {zonename}", cwd="ns3").out.strip()
        settime(
            f"-g HIDDEN -k OMNIPRESENT {TactN} -r OMNIPRESENT {TactN} -z OMNIPRESENT {TactN} -d OMNIPRESENT {TactN} {csk1_name}",
            cwd="ns3",
        )
        settime(
            f"-g OMNIPRESENT -k RUMOURED {TpubN1} -r RUMOURED {TpubN1} -z RUMOURED {TpubN1} -d HIDDEN {TpubN1} {csk2_name}",
            cwd="ns3",
        )
        # Signing.
        render_and_sign_zone(zonename, [csk1_name, csk2_name], extra_options="-z")

        # Step 3:
        # The zone signatures are also OMNIPRESENT.
        zonename = f"step3.{zone}"
        zones.append(Zone(zonename, f"{zonename}.db", Nameserver("ns3", "10.53.0.3")))
        isctest.log.info(f"setup {zonename}")
        # The time passed since the new algorithm keys have been introduced is 7 hours.
        TpubN1 = "now-7h"
        TsbmN1 = "now"
        csktimes = f"-P {TactN} -A {TactN}  -P sync {TsbmN} -I {TsbmN1}"
        newtimes = f"-P {TpubN1} -A {TpubN1} -P sync {TsbmN1}"
        # Key generation.
        csk1_name = keygen(f"-l csk1.conf {csktimes} {zonename}", cwd="ns3").out.strip()
        csk2_name = keygen(f"-l csk2.conf {newtimes} {zonename}", cwd="ns3").out.strip()
        settime(
            f"-g HIDDEN -k OMNIPRESENT {TactN} -r OMNIPRESENT {TactN} -z OMNIPRESENT {TactN} -d OMNIPRESENT {TactN} {csk1_name}",
            cwd="ns3",
        )
        settime(
            f"-g OMNIPRESENT -k OMNIPRESENT {TpubN1} -r OMNIPRESENT {TpubN1} -z RUMOURED {TpubN1} -d HIDDEN {TpubN1} {csk2_name}",
            cwd="ns3",
        )
        # Signing.
        render_and_sign_zone(zonename, [csk1_name, csk2_name], extra_options="-z")

        # Step 4:
        # The DS is swapped and can become OMNIPRESENT.
        zonename = f"step4.{zone}"
        zones.append(Zone(zonename, f"{zonename}.db", Nameserver("ns3", "10.53.0.3")))
        isctest.log.info(f"setup {zonename}")
        # The time passed since the DS has been swapped is 3 hours.
        TpubN1 = "now-10h"
        TsbmN1 = "now-3h"
        csktimes = f"-P {TactN} -A {TactN}  -P sync {TsbmN} -I {TsbmN1}"
        newtimes = f"-P {TpubN1} -A {TpubN1} -P sync {TsbmN1}"
        # Key generation.
        csk1_name = keygen(f"-l csk1.conf {csktimes} {zonename}", cwd="ns3").out.strip()
        csk2_name = keygen(f"-l csk2.conf {newtimes} {zonename}", cwd="ns3").out.strip()
        settime(
            f"-g HIDDEN -k OMNIPRESENT {TactN} -r OMNIPRESENT {TactN} -z OMNIPRESENT {TsbmN1} -d UNRETENTIVE {TsbmN1} -D ds {TsbmN1} {csk1_name}",
            cwd="ns3",
        )
        settime(
            f"-g OMNIPRESENT -k OMNIPRESENT {TpubN1} -r OMNIPRESENT {TpubN1} -z OMNIPRESENT {TsbmN1} -d RUMOURED {TsbmN1} -P ds {TsbmN1} {csk2_name}",
            cwd="ns3",
        )
        # Signing.
        render_and_sign_zone(zonename, [csk1_name, csk2_name], extra_options="-z")

        # Step 5:
        # The DNSKEY is removed long enough to be HIDDEN.
        zonename = f"step5.{zone}"
        zones.append(Zone(zonename, f"{zonename}.db", Nameserver("ns3", "10.53.0.3")))
        isctest.log.info(f"setup {zonename}")
        # The time passed since the DNSKEY has been removed is 2 hours.
        TpubN1 = "now-12h"
        TsbmN1 = "now-5h"
        csktimes = f"-P {TactN} -A {TactN} -P sync {TsbmN} -I {TsbmN1}"
        newtimes = f"-P {TpubN1} -A {TpubN1} -P sync {TsbmN1}"
        # Key generation.
        csk1_name = keygen(f"-l csk1.conf {csktimes} {zonename}", cwd="ns3").out.strip()
        csk2_name = keygen(f"-l csk2.conf {newtimes} {zonename}", cwd="ns3").out.strip()
        settime(
            f"-g HIDDEN -k UNRETENTIVE {TactN} -r UNRETENTIVE {TactN} -z UNRETENTIVE {TsbmN1} -d HIDDEN {TsbmN1} {csk1_name}",
            cwd="ns3",
        )
        settime(
            f"-g OMNIPRESENT -k OMNIPRESENT {TpubN1} -r OMNIPRESENT {TpubN1} -z OMNIPRESENT {TsbmN1} -d OMNIPRESENT {TsbmN1} {csk2_name}",
            cwd="ns3",
        )
        # Signing.
        render_and_sign_zone(zonename, [csk1_name, csk2_name], extra_options="-z")

        # Step 6:
        # The RRSIGs have been removed long enough to be HIDDEN.
        zonename = f"step6.{zone}"
        zones.append(Zone(zonename, f"{zonename}.db", Nameserver("ns3", "10.53.0.3")))
        isctest.log.info(f"setup {zonename}")
        # Additional time passed: 7h.
        TpubN1 = "now-19h"
        TsbmN1 = "now-12h"
        csktimes = f"-P {TactN}  -A {TactN}  -P sync {TsbmN} -I {TsbmN1}"
        newtimes = f"-P {TpubN1} -A {TpubN1} -P sync {TsbmN1}"
        # Key generation.
        csk1_name = keygen(f"-l csk1.conf {csktimes} {zonename}", cwd="ns3").out.strip()
        csk2_name = keygen(f"-l csk2.conf {newtimes} {zonename}", cwd="ns3").out.strip()
        settime(
            f"-g HIDDEN -k HIDDEN {TactN} -r UNRETENTIVE {TactN} -z UNRETENTIVE {TactN} -d HIDDEN {TsbmN1} {csk1_name}",
            cwd="ns3",
        )
        settime(
            f"-g OMNIPRESENT -k OMNIPRESENT {TpubN1} -r OMNIPRESENT {TpubN1} -z OMNIPRESENT {TsbmN1} -d OMNIPRESENT {TsbmN1} {csk2_name}",
            cwd="ns3",
        )
        # Signing.
        render_and_sign_zone(zonename, [csk1_name, csk2_name], extra_options="-z")

    return zones


def configure_algo_ksk_zsk(tld: str, reconfig: bool = False) -> List[Zone]:
    # The zones at algorithm-roll.$tld represent the various steps of a ZSK/KSK
    # algorithm rollover.
    zones = []
    zone = f"algorithm-roll.{tld}"
    keygen = EnvCmd("KEYGEN", "-L 3600")
    settime = EnvCmd("SETTIME", "-s")

    # Step 1:
    # Introduce the first key. This will immediately be active.
    zonename = f"step1.{zone}"
    zones.append(Zone(zonename, f"{zonename}.db", Nameserver("ns3", "10.53.0.3")))
    isctest.log.info(f"setup {zonename}")
    TactN = "now-7d"
    TsbmN = "now-161h"
    keytimes = f"-P {TactN} -A {TactN}"
    # Key generation.
    ksk_name = keygen(
        f"-a RSASHA256 -f KSK {keytimes} {zonename}", cwd="ns3"
    ).out.strip()
    zsk_name = keygen(f"-a RSASHA256 {keytimes} {zonename}", cwd="ns3").out.strip()
    settime(
        f"-g OMNIPRESENT -k OMNIPRESENT {TactN} -r OMNIPRESENT {TactN} -d OMNIPRESENT {TactN} {ksk_name}",
        cwd="ns3",
    )
    settime(
        f"-g OMNIPRESENT -k OMNIPRESENT {TactN} -z OMNIPRESENT {TactN} {zsk_name}",
        cwd="ns3",
    )
    # Signing.
    render_and_sign_zone(zonename, [ksk_name, zsk_name])

    if reconfig:
        # Step 2:
        # After the publication interval has passed the DNSKEY is OMNIPRESENT.
        zonename = f"step2.{zone}"
        zones.append(Zone(zonename, f"{zonename}.db", Nameserver("ns3", "10.53.0.3")))
        isctest.log.info(f"setup {zonename}")
        # The time passed since the new algorithm keys have been introduced is 3 hours.
        # Tsbm(N+1) = TpubN1 + Ipub = now + TTLsig + Dprp = now - 3h + 6h + 1h = now + 4h
        TpubN1 = "now-3h"
        TsbmN1 = "now+4h"
        ksk1times = f"-P {TactN}  -A {TactN}  -P sync {TsbmN} -I {TsbmN1}"
        zsk1times = f"-P {TactN}  -A {TactN}  -I {TsbmN1}"
        ksk2times = f"-P {TpubN1} -A {TpubN1} -P sync {TsbmN1}"
        zsk2times = f"-P {TpubN1} -A {TpubN1}"
        # Key generation.
        ksk1_name = keygen(
            f"-a RSASHA256 -f KSK {ksk1times} {zonename}", cwd="ns3"
        ).out.strip()
        zsk1_name = keygen(
            f"-a RSASHA256 {zsk1times} {zonename}", cwd="ns3"
        ).out.strip()
        ksk2_name = keygen(
            f"-a ECDSA256 -f KSK {ksk2times} {zonename}", cwd="ns3"
        ).out.strip()
        zsk2_name = keygen(f"-a ECDSA256 {zsk2times} {zonename}", cwd="ns3").out.strip()
        settime(
            f"-g HIDDEN -k OMNIPRESENT {TactN} -r OMNIPRESENT {TactN} -d OMNIPRESENT {TactN} {ksk1_name}",
            cwd="ns3",
        )
        settime(
            f"-g HIDDEN -k OMNIPRESENT {TactN} -z OMNIPRESENT {TactN} {zsk1_name}",
            cwd="ns3",
        )
        settime(
            f"-g OMNIPRESENT -k RUMOURED {TpubN1} -r RUMOURED {TpubN1} -d HIDDEN {TpubN1} {ksk2_name}",
            cwd="ns3",
        )
        settime(
            f"-g OMNIPRESENT -k RUMOURED {TpubN1} -z RUMOURED {TpubN1} {zsk2_name}",
            cwd="ns3",
        )
        # Signing.
        fake_lifetime(ksk1_name, 0)
        fake_lifetime(zsk1_name, 0)
        render_and_sign_zone(zonename, [ksk1_name, zsk1_name, ksk2_name, zsk2_name])

        # Step 3:
        # The zone signatures are also OMNIPRESENT.
        zonename = f"step3.{zone}"
        zones.append(Zone(zonename, f"{zonename}.db", Nameserver("ns3", "10.53.0.3")))
        isctest.log.info(f"setup {zonename}")
        # The time passed since the new algorithm keys have been introduced is 7 hours.
        TpubN1 = "now-7h"
        TsbmN1 = "now"
        ksk1times = f"-P {TactN} -A {TactN} -P sync {TsbmN} -I {TsbmN1}"
        zsk1times = f"-P {TactN} -A {TactN} -I {TsbmN1}"
        ksk2times = f"-P {TpubN1} -A {TpubN1} -P sync {TsbmN1}"
        zsk2times = f"-P {TpubN1} -A {TpubN1}"
        # Key generation.
        ksk1_name = keygen(
            f"-a RSASHA256 -f KSK {ksk1times} {zonename}", cwd="ns3"
        ).out.strip()
        zsk1_name = keygen(
            f"-a RSASHA256 {zsk1times} {zonename}", cwd="ns3"
        ).out.strip()
        ksk2_name = keygen(
            f"-a ECDSA256 -f KSK {ksk2times} {zonename}", cwd="ns3"
        ).out.strip()
        zsk2_name = keygen(f"-a ECDSA256 {zsk2times} {zonename}", cwd="ns3").out.strip()
        settime(
            f"-g HIDDEN -k OMNIPRESENT {TactN} -r OMNIPRESENT {TactN} -d OMNIPRESENT {TactN} {ksk1_name}",
            cwd="ns3",
        )
        settime(
            f"-g HIDDEN -k OMNIPRESENT {TactN} -z OMNIPRESENT {TactN} {zsk1_name}",
            cwd="ns3",
        )
        settime(
            f"-g OMNIPRESENT -k OMNIPRESENT {TpubN1} -r OMNIPRESENT {TpubN1} -d HIDDEN {TpubN1} {ksk2_name}",
            cwd="ns3",
        )
        settime(
            f"-g OMNIPRESENT -k OMNIPRESENT {TpubN1} -z RUMOURED {TpubN1} {zsk2_name}",
            cwd="ns3",
        )
        # Signing.
        fake_lifetime(ksk1_name, 0)
        fake_lifetime(zsk1_name, 0)
        render_and_sign_zone(zonename, [ksk1_name, zsk1_name, ksk2_name, zsk2_name])

        # Step 4:
        # The DS is swapped and can become OMNIPRESENT.
        zonename = f"step4.{zone}"
        zones.append(Zone(zonename, f"{zonename}.db", Nameserver("ns3", "10.53.0.3")))
        isctest.log.info(f"setup {zonename}")
        # The time passed since the DS has been swapped is 3 hours.
        TpubN1 = "now-10h"
        TsbmN1 = "now-3h"
        ksk1times = f"-P {TactN} -A {TactN} -P sync {TsbmN} -I {TsbmN1}"
        zsk1times = f"-P {TactN} -A {TactN} -I {TsbmN1}"
        ksk2times = f"-P {TpubN1} -A {TpubN1} -P sync {TsbmN1}"
        zsk2times = f"-P {TpubN1} -A {TpubN1}"
        # Key generation.
        ksk1_name = keygen(
            f"-a RSASHA256 -f KSK {ksk1times} {zonename}", cwd="ns3"
        ).out.strip()
        zsk1_name = keygen(
            f"-a RSASHA256 {zsk1times} {zonename}", cwd="ns3"
        ).out.strip()
        ksk2_name = keygen(
            f"-a ECDSA256 -f KSK {ksk2times} {zonename}", cwd="ns3"
        ).out.strip()
        zsk2_name = keygen(f"-a ECDSA256 {zsk2times} {zonename}", cwd="ns3").out.strip()
        settime(
            f"-g HIDDEN -k OMNIPRESENT {TactN} -r OMNIPRESENT {TactN} -d UNRETENTIVE {TsbmN1} -D ds {TsbmN1} {ksk1_name}",
            cwd="ns3",
        )
        settime(
            f"-g HIDDEN -k OMNIPRESENT {TactN} -z OMNIPRESENT {TactN} {zsk1_name}",
            cwd="ns3",
        )
        settime(
            f"-g OMNIPRESENT -k OMNIPRESENT {TpubN1} -r OMNIPRESENT {TpubN1} -d RUMOURED {TsbmN1} -P ds {TsbmN1} {ksk2_name}",
            cwd="ns3",
        )
        settime(
            f"-g OMNIPRESENT -k OMNIPRESENT {TpubN1} -z RUMOURED {TpubN1} {zsk2_name}",
            cwd="ns3",
        )
        # Signing.
        fake_lifetime(ksk1_name, 0)
        fake_lifetime(zsk1_name, 0)
        render_and_sign_zone(zonename, [ksk1_name, zsk1_name, ksk2_name, zsk2_name])

        # Step 5:
        # The DNSKEY is removed long enough to be HIDDEN.
        zonename = f"step5.{zone}"
        zones.append(Zone(zonename, f"{zonename}.db", Nameserver("ns3", "10.53.0.3")))
        isctest.log.info(f"setup {zonename}")
        # The time passed since the DNSKEY has been removed is 2 hours.
        TpubN1 = "now-12h"
        TsbmN1 = "now-5h"
        ksk1times = f"-P {TactN} -A {TactN} -P sync {TsbmN} -I {TsbmN1}"
        zsk1times = f"-P {TactN} -A {TactN} -I {TsbmN1}"
        ksk2times = f"-P {TpubN1} -A {TpubN1} -P sync {TsbmN1}"
        zsk2times = f"-P {TpubN1} -A {TpubN1}"
        # Key generation.
        ksk1_name = keygen(
            f"-a RSASHA256 -f KSK {ksk1times} {zonename}", cwd="ns3"
        ).out.strip()
        zsk1_name = keygen(
            f"-a RSASHA256 {zsk1times} {zonename}", cwd="ns3"
        ).out.strip()
        ksk2_name = keygen(
            f"-a ECDSA256 -f KSK {ksk2times} {zonename}", cwd="ns3"
        ).out.strip()
        zsk2_name = keygen(f"-a ECDSA256 {zsk2times} {zonename}", cwd="ns3").out.strip()
        settime(
            f"-g HIDDEN -k UNRETENTIVE {TsbmN1} -r UNRETENTIVE {TsbmN1} -d HIDDEN {TsbmN1} {ksk1_name}",
            cwd="ns3",
        )
        settime(
            f"-g HIDDEN -k UNRETENTIVE {TsbmN1} -z UNRETENTIVE {TsbmN1} {zsk1_name}",
            cwd="ns3",
        )
        settime(
            f"-g OMNIPRESENT -k OMNIPRESENT {TpubN1} -r OMNIPRESENT {TpubN1} -d OMNIPRESENT {TsbmN1} {ksk2_name}",
            cwd="ns3",
        )
        settime(
            f"-g OMNIPRESENT -k OMNIPRESENT {TpubN1} -z RUMOURED {TpubN1} {zsk2_name}",
            cwd="ns3",
        )
        # Signing.
        fake_lifetime(ksk1_name, 0)
        fake_lifetime(zsk1_name, 0)
        render_and_sign_zone(zonename, [ksk1_name, zsk1_name, ksk2_name, zsk2_name])

        # Step 6:
        # The RRSIGs have been removed long enough to be HIDDEN.
        zonename = f"step6.{zone}"
        zones.append(Zone(zonename, f"{zonename}.db", Nameserver("ns3", "10.53.0.3")))
        isctest.log.info(f"setup {zonename}")
        # Additional time passed: 7h.
        TpubN1 = "now-19h"
        TsbmN1 = "now-12h"
        ksk1times = f"-P {TactN} -A {TactN} -P sync {TsbmN} -I {TsbmN1}"
        zsk1times = f"-P {TactN} -A {TactN} -I {TsbmN1}"
        ksk2times = f"-P {TpubN1} -A {TpubN1} -P sync {TsbmN1}"
        zsk2times = f"-P {TpubN1} -A {TpubN1}"
        ksk1_name = keygen(
            f"-a RSASHA256 -f KSK {ksk1times} {zonename}", cwd="ns3"
        ).out.strip()
        zsk1_name = keygen(
            f"-a RSASHA256 {zsk1times} {zonename}", cwd="ns3"
        ).out.strip()
        ksk2_name = keygen(
            f"-a ECDSA256 -f KSK {ksk2times} {zonename}", cwd="ns3"
        ).out.strip()
        zsk2_name = keygen(f"-a ECDSA256 {zsk2times} {zonename}", cwd="ns3").out.strip()
        settime(
            f"-g HIDDEN -k HIDDEN {TsbmN1} -r UNRETENTIVE {TsbmN1} -d HIDDEN {TsbmN1} {ksk1_name}",
            cwd="ns3",
        )
        settime(
            f"-g HIDDEN -k HIDDEN {TsbmN1} -z UNRETENTIVE {TsbmN1} {zsk1_name}",
            cwd="ns3",
        )
        settime(
            f"-g OMNIPRESENT -k OMNIPRESENT {TpubN1} -r OMNIPRESENT {TpubN1} -d OMNIPRESENT {TsbmN1} {ksk2_name}",
            cwd="ns3",
        )
        settime(
            f"-g OMNIPRESENT -k OMNIPRESENT {TpubN1} -z RUMOURED {TpubN1} {zsk2_name}",
            cwd="ns3",
        )
        # Signing.
        fake_lifetime(ksk1_name, 0)
        fake_lifetime(zsk1_name, 0)
        render_and_sign_zone(zonename, [ksk1_name, zsk1_name, ksk2_name, zsk2_name])

    return zones


def configure_cskroll1(tld: str, policy: str) -> List[Zone]:
    # The zones at csk-roll1.$tld represent the various steps of a CSK rollover
    # (which is essentially a ZSK Pre-Publication / KSK Double-KSK rollover).
    zones = []
    zone = f"csk-roll1.{tld}"
    cds = "cdnskey,cds:sha384"
    keygen = EnvCmd("KEYGEN", f"-k {policy} -l kasp.conf")
    settime = EnvCmd("SETTIME", "-s")

    # Step 1:
    # Introduce the first key. This will immediately be active.
    zonename = f"step1.{zone}"
    zones.append(Zone(zonename, f"{zonename}.db", Nameserver("ns3", "10.53.0.3")))
    isctest.log.info(f"setup {zonename}")
    TactN = "now-7d"
    keytimes = f"-P {TactN} -A {TactN}"
    # Key generation.
    csk_name = keygen(f"{keytimes} {zonename}", cwd="ns3").out.strip()
    settime(
        f"-g OMNIPRESENT -k OMNIPRESENT {TactN} -r OMNIPRESENT {TactN} -z OMNIPRESENT {TactN} -d OMNIPRESENT {TactN} {csk_name}",
        cwd="ns3",
    )
    # Signing.
    render_and_sign_zone(zonename, [csk_name], extra_options=f"-z -G {cds}")

    # Step 2:
    # It is time to introduce the new CSK.
    zonename = f"step2.{zone}"
    zones.append(Zone(zonename, f"{zonename}.db", Nameserver("ns3", "10.53.0.3")))
    isctest.log.info(f"setup {zonename}")
    # According to RFC 7583:
    # KSK: Tpub(N+1) <= Tact(N) + Lksk - IpubC
    # ZSK: Tpub(N+1) <= Tact(N) + Lzsk - Ipub
    # IpubC = DprpC + TTLkey (+publish-safety)
    # Ipub  = IpubC
    # Lcsk = Lksk = Lzsk
    #
    # Lcsk:           6mo (186d, 4464h)
    # Dreg:           N/A
    # DprpC:          1h
    # TTLkey:         1h
    # publish-safety: 1h
    # Ipub:           3h
    #
    # Tact(N) = now - Lcsk + Ipub = now - 186d + 3h
    #         = now - 4464h + 3h  = now - 4461h
    TactN = "now-4461h"
    keytimes = f"-P {TactN} -A {TactN}"
    # Key generation.
    csk_name = keygen(f"{keytimes} {zonename}", cwd="ns3").out.strip()
    settime(
        f"-g OMNIPRESENT -k OMNIPRESENT {TactN} -r OMNIPRESENT {TactN} -z OMNIPRESENT {TactN} -d OMNIPRESENT {TactN} {csk_name}",
        cwd="ns3",
    )
    # Signing.
    render_and_sign_zone(zonename, [csk_name], extra_options=f"-z -G {cds}")

    # Step 3:
    # It is time to submit the DS and to roll signatures.
    zonename = f"step3.{zone}"
    zones.append(Zone(zonename, f"{zonename}.db", Nameserver("ns3", "10.53.0.3")))
    isctest.log.info(f"setup {zonename}")
    # According to RFC 7583:
    #
    # Tsbm(N+1) >= Trdy(N+1)
    # KSK: Tact(N+1) = Tsbm(N+1)
    # ZSK: Tact(N+1) = Tpub(N+1) + Ipub = Tsbm(N+1)
    # KSK: Iret  = DprpP + TTLds (+retire-safety)
    # ZSK: IretZ = Dsgn + Dprp + TTLsig (+retire-safety)
    #
    # Lcsk:           186d
    # Dprp:           1h
    # DprpP:          1h
    # Dreg:           N/A
    # Dsgn:           25d
    # TTLds:          1h
    # TTLsig:         1d
    # retire-safety:  2h
    # Iret:           4h
    # IretZ:          26d3h
    # Ipub:           3h
    #
    # Tpub(N)   = now - Lcsk = now - 186d
    # Tact(N)   = now - Lcsk + Dprp + TTLsig = now - 4439h
    # Tret(N)   = now
    # Trem(N)   = now + IretZ = now + 26d3h = now + 627h
    # Tpub(N+1) = now - Ipub = now - 3h
    # Tact(N+1) = Tret(N)
    # Tret(N+1) = now + Lcsk = now + 186d = now + 186d
    # Trem(N+1) = now + Lcsk + IretZ = now + 186d + 26d3h =
    #           = now + 5091h
    TpubN = "now-186d"
    TactN = "now-4439h"
    TretN = "now"
    TremN = "now+627h"
    TpubN1 = "now-3h"
    TactN1 = TretN
    TretN1 = "now+186d"
    TremN1 = "now+5091h"
    keytimes = (
        f"-P {TpubN} -P sync {TactN} -A {TpubN} -I {TretN} -D {TremN} -D sync {TactN1}"
    )
    newtimes = f"-P {TpubN1} -P sync {TactN1} -A {TactN1} -I {TretN1} -D {TremN1}"
    # Key generation.
    csk1_name = keygen(f"{keytimes} {zonename}", cwd="ns3").out.strip()
    csk2_name = keygen(f"{newtimes} {zonename}", cwd="ns3").out.strip()
    settime(
        f"-g HIDDEN -k OMNIPRESENT {TactN} -r OMNIPRESENT {TactN} -z OMNIPRESENT {TactN} -d OMNIPRESENT {TactN} {csk1_name}",
        cwd="ns3",
    )
    settime(
        f"-g OMNIPRESENT -k RUMOURED {TpubN1} -r RUMOURED {TpubN1} -z HIDDEN {TpubN1} -d HIDDEN {TpubN1} {csk2_name}",
        cwd="ns3",
    )
    # Set key rollover relationship.
    set_key_relationship(csk1_name, csk2_name)
    # Signing.
    render_and_sign_zone(zonename, [csk1_name, csk2_name], extra_options=f"-z -G {cds}")

    # Step 4:
    # Some time later all the ZRRSIG records should be from the new CSK, and the
    # DS should be swapped.  The ZRRSIG records are all replaced after IretZ
    # (which is 26d3h).  The DS is swapped after Iret (which is 4h).
    # In other words, the DS is swapped before all zone signatures are replaced.
    zonename = f"step4.{zone}"
    zones.append(Zone(zonename, f"{zonename}.db", Nameserver("ns3", "10.53.0.3")))
    isctest.log.info(f"setup {zonename}")
    # According to RFC 7583:
    # Trem(N)    = Tret(N) - Iret + IretZ
    # now       = Tsbm(N+1) + Iret
    #
    # Lcsk:   186d
    # Iret:   4h
    # IretZ:  26d3h
    #
    # Tpub(N)   = now - Iret - Lcsk = now - 4h - 186d = now - 4468h
    # Tret(N)   = now - Iret = now - 4h = now - 4h
    # Trem(N)   = now - Iret + IretZ = now - 4h + 26d3h
    #           = now + 623h
    # Tpub(N+1) = now - Iret - IpubC = now - 4h - 3h = now - 7h
    # Tact(N+1) = Tret(N)
    # Tret(N+1) = now - Iret + Lcsk = now - 4h + 186d = now + 4460h
    # Trem(N+1) = now - Iret + Lcsk + IretZ = now - 4h + 186d + 26d3h
    #           = now + 5087h
    TpubN = "now-4468h"
    TactN = "now-4443h"
    TretN = "now-4h"
    TremN = "now+623h"
    TpubN1 = "now-7h"
    TactN1 = TretN
    TretN1 = "now+4460h"
    TremN1 = "now+5087h"
    keytimes = (
        f"-P {TpubN} -P sync {TactN} -A {TpubN} -I {TretN} -D {TremN} -D sync {TactN1}"
    )
    newtimes = f"-P {TpubN1} -P sync {TactN1} -A {TactN1} -I {TretN1} -D {TremN1}"
    # Key generation.
    csk1_name = keygen(f"{keytimes} {zonename}", cwd="ns3").out.strip()
    csk2_name = keygen(f"{newtimes} {zonename}", cwd="ns3").out.strip()
    settime(
        f"-g HIDDEN -k OMNIPRESENT {TactN} -r OMNIPRESENT {TactN} -z UNRETENTIVE {TactN1} -d UNRETENTIVE {TactN1} -D ds {TactN1} {csk1_name}",
        cwd="ns3",
    )
    settime(
        f"-g OMNIPRESENT -k OMNIPRESENT {TactN1} -r OMNIPRESENT {TactN1} -z RUMOURED {TactN1} -d RUMOURED {TactN1} -P ds {TactN1} {csk2_name}",
        cwd="ns3",
    )
    # Set key rollover relationship.
    set_key_relationship(csk1_name, csk2_name)
    # Signing.
    render_and_sign_zone(zonename, [csk1_name, csk2_name], extra_options=f"-z -G {cds}")

    # Step 5:
    # After the DS is swapped in step 4, also the KRRSIG records can be removed.
    # At this time these have all become hidden.
    zonename = f"step5.{zone}"
    zones.append(Zone(zonename, f"{zonename}.db", Nameserver("ns3", "10.53.0.3")))
    isctest.log.info(f"setup {zonename}")
    # Subtract DNSKEY TTL plus zone propagation delay from all the times (2h).
    TpubN = "now-4470h"
    TactN = "now-4445h"
    TretN = "now-6h"
    TremN = "now+621h"
    TpubN1 = "now-9h"
    TactN1 = TretN
    TretN1 = "now+4458h"
    TremN1 = "now+5085h"
    keytimes = (
        f"-P {TpubN} -P sync {TactN} -A {TpubN} -I {TretN} -D {TremN} -D sync {TactN1}"
    )
    newtimes = f"-P {TpubN1} -P sync {TactN1} -A {TactN1} -I {TretN1} -D {TremN1}"
    # Key generation.
    csk1_name = keygen(f"{keytimes} {zonename}", cwd="ns3").out.strip()
    csk2_name = keygen(f"{newtimes} {zonename}", cwd="ns3").out.strip()
    settime(
        f"-g HIDDEN -k OMNIPRESENT {TactN} -r UNRETENTIVE now-2h -z UNRETENTIVE {TactN1} -d HIDDEN now-2h {csk1_name}",
        cwd="ns3",
    )
    settime(
        f"-g OMNIPRESENT -k OMNIPRESENT {TactN1} -r OMNIPRESENT {TactN1} -z RUMOURED {TactN1} -d OMNIPRESENT now-2h {csk2_name}",
        cwd="ns3",
    )
    # Set key rollover relationship.
    set_key_relationship(csk1_name, csk2_name)
    # Signing.
    render_and_sign_zone(zonename, [csk1_name, csk2_name], extra_options=f"-z -G {cds}")

    # Step 6:
    # After the retire interval has passed the predecessor DNSKEY can be
    # removed from the zone.
    zonename = f"step6.{zone}"
    zones.append(Zone(zonename, f"{zonename}.db", Nameserver("ns3", "10.53.0.3")))
    isctest.log.info(f"setup {zonename}")
    # According to RFC 7583:
    # Trem(N) = Tret(N) + IretZ
    # Tret(N) = Tact(N) + Lcsk
    #
    # Lcsk:   186d
    # Iret:   4h
    # IretZ:  26d3h
    #
    # Tpub(N)   = now - IretZ - Lcsk = now - 627h - 186d
    #           = now - 627h - 4464h = now - 5091h
    # Tact(N)   = now - 627h - 186d
    # Tret(N)   = now - IretZ = now - 627h
    # Trem(N)   = now
    # Tpub(N+1) = now - IretZ - Ipub = now - 627h - 3h = now - 630h
    # Tact(N+1) = Tret(N)
    # Tret(N+1) = now - IretZ + Lcsk = now - 627h + 186d = now + 3837h
    # Trem(N+1) = now + Lcsk = now + 186d
    TpubN = "now-5091h"
    TactN = "now-5066h"
    TretN = "now-627h"
    TremN = "now"
    TpubN1 = "now-630h"
    TactN1 = TretN
    TretN1 = "now+3837h"
    TremN1 = "now+186d"
    keytimes = (
        f"-P {TpubN} -P sync {TactN} -A {TpubN} -I {TretN} -D {TremN} -D sync {TactN1}"
    )
    newtimes = f"-P {TpubN1} -P sync {TactN1} -A {TactN1} -I {TretN1} -D {TremN1}"
    # Key generation.
    csk1_name = keygen(f"{keytimes} {zonename}", cwd="ns3").out.strip()
    csk2_name = keygen(f"{newtimes} {zonename}", cwd="ns3").out.strip()
    settime(
        f"-g HIDDEN -k OMNIPRESENT {TactN} -r HIDDEN {TremN} -z UNRETENTIVE {TactN1} -d HIDDEN {TremN} {csk1_name}",
        cwd="ns3",
    )
    settime(
        f"-g OMNIPRESENT -k OMNIPRESENT {TactN1} -r OMNIPRESENT {TactN1} -z RUMOURED {TactN1} -d OMNIPRESENT {TremN} {csk2_name}",
        cwd="ns3",
    )
    # Set key rollover relationship.
    set_key_relationship(csk1_name, csk2_name)
    # Signing.
    render_and_sign_zone(zonename, [csk1_name, csk2_name], extra_options=f"-z -G {cds}")

    # Step 7:
    # Some time later the predecessor DNSKEY enters the HIDDEN state.
    zonename = f"step7.{zone}"
    zones.append(Zone(zonename, f"{zonename}.db", Nameserver("ns3", "10.53.0.3")))
    isctest.log.info(f"setup {zonename}")
    # Subtract DNSKEY TTL plus zone propagation delay from all the times (2h).
    TpubN = "now-5093h"
    TactN = "now-5068h"
    TretN = "now-629h"
    TremN = "now-2h"
    TpubN1 = "now-632h"
    TactN1 = TretN
    TretN1 = "now+3835h"
    TremN1 = "now+4462h"
    keytimes = (
        f"-P {TpubN} -P sync {TactN} -A {TpubN} -I {TretN} -D {TremN} -D sync {TactN1}"
    )
    newtimes = f"-P {TpubN1} -P sync {TactN1} -A {TactN1} -I {TretN1} -D {TremN1}"
    # Key generation.
    csk1_name = keygen(f"{keytimes} {zonename}", cwd="ns3").out.strip()
    csk2_name = keygen(f"{newtimes} {zonename}", cwd="ns3").out.strip()
    settime(
        f"-g HIDDEN -k UNRETENTIVE {TremN} -r HIDDEN {TremN} -z HIDDEN {TactN1} -d HIDDEN {TremN} {csk1_name}",
        cwd="ns3",
    )
    settime(
        f"-g OMNIPRESENT -k OMNIPRESENT {TactN1} -r OMNIPRESENT {TactN1} -z OMNIPRESENT {TactN1} -d OMNIPRESENT {TactN1} {csk2_name}",
        cwd="ns3",
    )
    # Set key rollover relationship.
    set_key_relationship(csk1_name, csk2_name)
    # Signing.
    render_and_sign_zone(zonename, [csk1_name, csk2_name], extra_options=f"-z -G {cds}")

    # Step 8:
    # The predecessor DNSKEY can be purged.
    zonename = f"step8.{zone}"
    zones.append(Zone(zonename, f"{zonename}.db", Nameserver("ns3", "10.53.0.3")))
    isctest.log.info(f"setup {zonename}")
    # Subtract purge-keys interval from all the times (1h).
    TpubN = "now-5094h"
    TactN = "now-5069h"
    TretN = "now-630h"
    TremN = "now-3h"
    TpubN1 = "now-633h"
    TactN1 = TretN
    TretN1 = "now+3834h"
    TremN1 = "now+4461h"
    keytimes = (
        f"-P {TpubN} -P sync {TactN} -A {TpubN} -I {TretN} -D {TremN} -D sync {TactN1}"
    )
    newtimes = f"-P {TpubN1} -P sync {TactN1} -A {TactN1} -I {TretN1} -D {TremN1}"
    # Key generation.
    csk1_name = keygen(f"{keytimes} {zonename}", cwd="ns3").out.strip()
    csk2_name = keygen(f"{newtimes} {zonename}", cwd="ns3").out.strip()
    settime(
        f"-g HIDDEN -k HIDDEN {TremN} -r HIDDEN {TremN} -z HIDDEN {TactN1} -d HIDDEN {TremN} {csk1_name}",
        cwd="ns3",
    )
    settime(
        f"-g OMNIPRESENT -k OMNIPRESENT {TactN1} -r OMNIPRESENT {TactN1} -z OMNIPRESENT {TactN1} -d OMNIPRESENT {TactN1} {csk2_name}",
        cwd="ns3",
    )
    # Set key rollover relationship.
    set_key_relationship(csk1_name, csk2_name)
    # Signing.
    render_and_sign_zone(zonename, [csk1_name, csk2_name], extra_options=f"-z -G {cds}")

    return zones


def configure_cskroll2(tld: str, policy: str) -> List[Zone]:
    # The zones at csk-roll2.$tld represent the various steps of a CSK rollover
    # (which is essentially a ZSK Pre-Publication / KSK Double-KSK rollover).
    # This scenario differs from the csk-roll1 one because the zone signatures (ZRRSIG)
    # are replaced with the new key sooner than the DS is swapped.
    zones = []
    zone = f"csk-roll2.{tld}"
    cds = "cdnskey,cds:sha-256,cds:sha-384"
    keygen = EnvCmd("KEYGEN", f"-k {policy} -l kasp.conf")
    settime = EnvCmd("SETTIME", "-s")

    # Step 1:
    # Introduce the first key. This will immediately be active.
    zonename = f"step1.{zone}"
    zones.append(Zone(zonename, f"{zonename}.db", Nameserver("ns3", "10.53.0.3")))
    isctest.log.info(f"setup {zonename}")
    TactN = "now-7d"
    keytimes = f"-P {TactN} -A {TactN}"
    # Key generation.
    csk_name = keygen(f"{keytimes} {zonename}", cwd="ns3").out.strip()
    settime(
        f"-g OMNIPRESENT -k OMNIPRESENT {TactN} -r OMNIPRESENT {TactN} -z OMNIPRESENT {TactN} -d OMNIPRESENT {TactN} {csk_name}",
        cwd="ns3",
    )
    # Signing.
    render_and_sign_zone(zonename, [csk_name], extra_options=f"-z -G {cds}")

    # Step 2:
    # It is time to introduce the new CSK.
    zonename = f"step2.{zone}"
    zones.append(Zone(zonename, f"{zonename}.db", Nameserver("ns3", "10.53.0.3")))
    isctest.log.info(f"setup {zonename}")
    # According to RFC 7583:
    # KSK: Tpub(N+1) <= Tact(N) + Lksk - IpubC
    # ZSK: Tpub(N+1) <= Tact(N) + Lzsk - Ipub
    # IpubC = DprpC + TTLkey (+publish-safety)
    # Ipub  = IpubC
    # Lcsk = Lksk = Lzsk
    #
    # Lcsk:           6mo (186d, 4464h)
    # Dreg:           N/A
    # DprpC:          1h
    # TTLkey:         1h
    # publish-safety: 1h
    # Ipub:           3h
    #
    # Tact(N)  = now - Lcsk + Ipub = now - 186d + 3h
    #          = now - 4464h + 3h = now - 4461h
    TactN = "now-4461h"
    keytimes = f"-P {TactN} -A {TactN}"
    # Key generation.
    csk_name = keygen(f"{keytimes} {zonename}", cwd="ns3").out.strip()
    settime(
        f"-g OMNIPRESENT -k OMNIPRESENT {TactN} -r OMNIPRESENT {TactN} -z OMNIPRESENT {TactN} -d OMNIPRESENT {TactN} {csk_name}",
        cwd="ns3",
    )
    # Signing.
    render_and_sign_zone(zonename, [csk_name], extra_options=f"-z -G {cds}")

    # Step 3:
    # It is time to submit the DS and to roll signatures.
    zonename = f"step3.{zone}"
    zones.append(Zone(zonename, f"{zonename}.db", Nameserver("ns3", "10.53.0.3")))
    isctest.log.info(f"setup {zonename}")
    # According to RFC 7583:
    #
    # Tsbm(N+1) >= Trdy(N+1)
    # KSK: Tact(N+1) = Tsbm(N+1)
    # ZSK: Tact(N+1) = Tpub(N+1) + Ipub = Tsbm(N+1)
    # KSK: Iret  = DprpP + TTLds (+retire-safety)
    # ZSK: IretZ = Dsgn + Dprp + TTLsig (+retire-safety)
    #
    # Lcsk:           186d
    # Dprp:           1h
    # DprpP:          1w
    # Dreg:           N/A
    # Dsgn:           12h
    # TTLds:          1h
    # TTLsig:         1d
    # retire-safety:  1h
    # Iret:           170h
    # IretZ:          38h
    # Ipub:           3h
    #
    # Tpub(N)   = now - Lcsk = now - 186d
    # Tact(N)   = now - Lcsk + Dprp + TTLsig = now - 4439h
    # Tret(N)   = now
    # Trem(N)   = now + IretZ = now + 26d3h = now + 627h
    # Tpub(N+1) = now - Ipub = now - 3h
    # Tact(N+1) = Tret(N)
    # Tret(N+1) = now + Lcsk = now + 186d = now + 186d
    # Trem(N+1) = now + Lcsk + IretZ = now + 186d + 26d3h =
    #           = now + 5091h
    TpubN = "now-186d"
    TactN = "now-4439h"
    TretN = "now"
    TremN = "now+170h"
    TpubN1 = "now-3h"
    TactN1 = TretN
    TretN1 = "now+186d"
    TremN1 = "now+4634h"
    keytimes = (
        f"-P {TpubN} -P sync {TactN} -A {TpubN} -I {TretN} -D {TremN} -D sync {TactN1}"
    )
    newtimes = f"-P {TpubN1} -P sync {TactN1} -A {TactN1} -I {TretN1} -D {TremN1}"
    # Key generation.
    csk1_name = keygen(f"{keytimes} {zonename}", cwd="ns3").out.strip()
    csk2_name = keygen(f"{newtimes} {zonename}", cwd="ns3").out.strip()
    settime(
        f"-g HIDDEN -k OMNIPRESENT {TactN} -r OMNIPRESENT {TactN} -z OMNIPRESENT {TactN} -d OMNIPRESENT {TactN} {csk1_name}",
        cwd="ns3",
    )
    settime(
        f"-g OMNIPRESENT -k RUMOURED {TpubN1} -r RUMOURED {TpubN1} -z HIDDEN {TpubN1} -d HIDDEN {TpubN1} {csk2_name}",
        cwd="ns3",
    )
    # Set key rollover relationship.
    set_key_relationship(csk1_name, csk2_name)
    # Signing.
    render_and_sign_zone(zonename, [csk1_name, csk2_name], extra_options=f"-z -G {cds}")

    # Step 4:
    # Some time later all the ZRRSIG records should be from the new CSK, and the
    # DS should be swapped.  The ZRRSIG records are all replaced after IretZ (38h).
    # The DS is swapped after Dreg + Iret (1w3h). In other words, the zone
    # signatures are replaced before the DS is swapped.
    zonename = f"step4.{zone}"
    zones.append(Zone(zonename, f"{zonename}.db", Nameserver("ns3", "10.53.0.3")))
    isctest.log.info(f"setup {zonename}")
    # According to RFC 7583:
    # Trem(N)    = Tret(N) + IretZ
    #
    # Lcsk:   186d
    # Dreg:   N/A
    # Iret:   170h
    # IretZ:  38h
    #
    # Tpub(N)    = now - IretZ - Lcsk = now - 38h - 186d
    #            = now - 38h - 4464h = now - 4502h
    # Tact(N)    = now - Iret - Lcsk + TTLsig = now - 4502h + 25h = now - 4477h
    # Tret(N)    = now - IretZ = now - 38h
    # Trem(N)    = now - IretZ + Iret = now - 38h + 170h = now + 132h
    # Tpub(N+1)  = now - IretZ - IpubC = now - 38h - 3h = now - 41h
    # Tact(N+1)  = Tret(N)
    # Tret(N+1)  = now - IretZ + Lcsk = now - 38h + 186d
    #            = now + 4426h
    # Trem(N+1)  = now - IretZ + Lcsk + Iret
    #            = now + 4426h + 3h = now + 4429h
    TpubN = "now-4502h"
    TactN = "now-4477h"
    TretN = "now-38h"
    TremN = "now+132h"
    TpubN1 = "now-41h"
    TactN1 = TretN
    TretN1 = "now+4426h"
    TremN1 = "now+4429h"
    keytimes = (
        f"-P {TpubN} -P sync {TactN} -A {TpubN} -I {TretN} -D {TremN} -D sync {TactN1}"
    )
    newtimes = f"-P {TpubN1} -P sync {TactN1} -A {TactN1} -I {TretN1} -D {TremN1}"
    # Key generation.
    csk1_name = keygen(f"{keytimes} {zonename}", cwd="ns3").out.strip()
    csk2_name = keygen(f"{newtimes} {zonename}", cwd="ns3").out.strip()
    settime(
        f"-g HIDDEN -k OMNIPRESENT {TactN} -r OMNIPRESENT {TactN} -z UNRETENTIVE {TretN} -d UNRETENTIVE {TretN} -D ds {TretN} {csk1_name}",
        cwd="ns3",
    )
    settime(
        f"-g OMNIPRESENT -k OMNIPRESENT {TactN1} -r OMNIPRESENT {TactN1} -z RUMOURED {TactN1} -d RUMOURED {TactN1} -P ds {TactN1} {csk2_name}",
        cwd="ns3",
    )
    # Set key rollover relationship.
    set_key_relationship(csk1_name, csk2_name)
    # Signing.
    render_and_sign_zone(zonename, [csk1_name, csk2_name], extra_options=f"-z -G {cds}")

    # Step 5:
    # Some time later the DS can be swapped and the old DNSKEY can be removed from
    # the zone.
    zonename = f"step5.{zone}"
    zones.append(Zone(zonename, f"{zonename}.db", Nameserver("ns3", "10.53.0.3")))
    isctest.log.info(f"setup {zonename}")
    # Subtract Iret (170h) - IretZ (38h) = 132h.
    #
    # Tpub(N)   = now - 4502h - 132h = now - 4634h
    # Tact(N)   = now - 4477h - 132h = now - 4609h
    # Tret(N)   = now - 38h - 132h = now - 170h
    # Trem(N)   = now + 132h - 132h = now
    # Tpub(N+1) = now - 41h - 132h = now - 173h
    # Tact(N+1) = Tret(N)
    # Tret(N+1) = now + 4426h - 132h = now + 4294h
    # Trem(N+1) = now + 4492h - 132h = now + 4360h
    TpubN = "now-4634h"
    TactN = "now-4609h"
    TretN = "now-170h"
    TremN = "now"
    TpubN1 = "now-173h"
    TactN1 = TretN
    TretN1 = "now+4294h"
    TremN1 = "now+4360h"
    keytimes = (
        f"-P {TpubN} -P sync {TactN} -A {TpubN} -I {TretN} -D {TremN} -D sync {TactN1}"
    )
    newtimes = f"-P {TpubN1} -P sync {TactN1} -A {TactN1} -I {TretN1} -D {TremN1}"
    # Key generation.
    csk1_name = keygen(f"{keytimes} {zonename}", cwd="ns3").out.strip()
    csk2_name = keygen(f"{newtimes} {zonename}", cwd="ns3").out.strip()
    settime(
        f"-g HIDDEN -k OMNIPRESENT {TactN} -r OMNIPRESENT {TactN} -z HIDDEN now-133h -d UNRETENTIVE {TactN1} -D ds {TactN1} {csk1_name}",
        cwd="ns3",
    )
    settime(
        f"-g OMNIPRESENT -k OMNIPRESENT {TactN1} -r OMNIPRESENT {TactN1} -z OMNIPRESENT now-133h -d RUMOURED {TactN1} -P ds {TactN1} {csk2_name}",
        cwd="ns3",
    )
    # Set key rollover relationship.
    set_key_relationship(csk1_name, csk2_name)
    # Signing.
    render_and_sign_zone(zonename, [csk1_name, csk2_name], extra_options=f"-z -G {cds}")

    # Step 6:
    # Some time later the predecessor DNSKEY enters the HIDDEN state.
    zonename = f"step6.{zone}"
    zones.append(Zone(zonename, f"{zonename}.db", Nameserver("ns3", "10.53.0.3")))
    isctest.log.info(f"setup {zonename}")
    # Subtract DNSKEY TTL plus zone propagation delay (2h).
    #
    # Tpub(N)   = now - 4634h - 2h = now - 4636h
    # Tact(N)   = now - 4609h - 2h = now - 4611h
    # Tret(N)   = now - 170h - 2h = now - 172h
    # Trem(N)   = now - 2h
    # Tpub(N+1) = now - 173h - 2h = now - 175h
    # Tact(N+1) = Tret(N)
    # Tret(N+1) = now + 4294h - 2h = now + 4292h
    # Trem(N+1) = now + 4360h - 2h = now + 4358h
    TpubN = "now-4636h"
    TactN = "now-4611h"
    TretN = "now-172h"
    TremN = "now-2h"
    TpubN1 = "now-175h"
    TactN1 = TretN
    TretN1 = "now+4292h"
    TremN1 = "now+4358h"
    keytimes = (
        f"-P {TpubN} -P sync {TactN} -A {TpubN} -I {TretN} -D {TremN} -D sync {TactN1}"
    )
    newtimes = f"-P {TpubN1} -P sync {TactN1} -A {TactN1} -I {TretN1} -D {TremN1}"
    # Key generation.
    csk1_name = keygen(f"{keytimes} {zonename}", cwd="ns3").out.strip()
    csk2_name = keygen(f"{newtimes} {zonename}", cwd="ns3").out.strip()
    settime(
        f"-g HIDDEN -k UNRETENTIVE {TremN} -r UNRETENTIVE {TremN} -z HIDDEN now-135h -d HIDDEN {TremN} {csk1_name}",
        cwd="ns3",
    )
    settime(
        f"-g OMNIPRESENT -k OMNIPRESENT {TactN1} -r OMNIPRESENT {TactN1} -z OMNIPRESENT now-135h -d OMNIPRESENT {TremN} {csk2_name}",
        cwd="ns3",
    )
    # Set key rollover relationship.
    set_key_relationship(csk1_name, csk2_name)
    # Signing.
    render_and_sign_zone(zonename, [csk1_name, csk2_name], extra_options=f"-z -G {cds}")

    # Step 7:
    # The predecessor DNSKEY can be purged, but purge-keys is disabled.
    zonename = f"step7.{zone}"
    zones.append(Zone(zonename, f"{zonename}.db", Nameserver("ns3", "10.53.0.3")))
    isctest.log.info(f"setup {zonename}")
    # Subtract 90 days (default, 2160h) from all the times.
    #
    # Tpub(N)   = now - 4636h - 2160h = now - 6796h
    # Tact(N)   = now - 4611h - 2160h = now - 6771h
    # Tret(N)   = now - 172h - 2160h = now - 2332h
    # Trem(N)   = now - 2h - 2160h = now - 2162h
    # Tpub(N+1) = now - 175h - 2160h = now - 2335h
    # Tact(N+1) = Tret(N)
    # Tret(N+1) = now + 4292h - 2160h = now + 2132h
    # Trem(N+1) = now + 4358h - 2160h = now + 2198h
    TpubN = "now-6796h"
    TactN = "now-6771h"
    TretN = "now-2332h"
    TremN = "now-2162h"
    TpubN1 = "now-2335h"
    TactN1 = TretN
    TretN1 = "now+2132h"
    TremN1 = "now+2198h"

    keytimes = (
        f"-P {TpubN} -P sync {TactN} -A {TpubN} -I {TretN} -D {TremN} -D sync {TactN1}"
    )
    newtimes = f"-P {TpubN1} -P sync {TactN1} -A {TactN1} -I {TretN1} -D {TremN1}"
    # Key generation.
    csk1_name = keygen(f"{keytimes} {zonename}", cwd="ns3").out.strip()
    csk2_name = keygen(f"{newtimes} {zonename}", cwd="ns3").out.strip()
    settime(
        f"-g HIDDEN -k UNRETENTIVE {TremN} -r HIDDEN {TremN} -z HIDDEN {TactN1} -d HIDDEN {TremN} {csk1_name}",
        cwd="ns3",
    )
    settime(
        f"-g OMNIPRESENT -k OMNIPRESENT {TactN1} -r OMNIPRESENT {TactN1} -z OMNIPRESENT {TactN1} -d OMNIPRESENT {TremN} {csk2_name}",
        cwd="ns3",
    )
    # Set key rollover relationship.
    set_key_relationship(csk1_name, csk2_name)
    # Signing.
    render_and_sign_zone(zonename, [csk1_name, csk2_name], extra_options=f"-z -G {cds}")

    # Step 8:
    # The predecessor DNSKEY can be purged.
    zonename = f"step8.{zone}"
    zones.append(Zone(zonename, f"{zonename}.db", Nameserver("ns3", "10.53.0.3")))
    isctest.log.info(f"setup {zonename}")
    # Subtract purge-keys interval from all the times (1h).
    TpubN = "now-5094h"
    TactN = "now-5069h"
    TretN = "now-630h"
    TremN = "now-3h"
    TpubN1 = "now-633h"
    TactN1 = TretN
    TretN1 = "now+3834h"
    TremN1 = "now+4461h"
    keytimes = (
        f"-P {TpubN} -P sync {TactN} -A {TpubN} -I {TretN} -D {TremN} -D sync {TactN1}"
    )
    newtimes = f"-P {TpubN1} -P sync {TactN1} -A {TactN1} -I {TretN1} -D {TremN1}"
    # Key generation.
    csk1_name = keygen(f"{keytimes} {zonename}", cwd="ns3").out.strip()
    csk2_name = keygen(f"{newtimes} {zonename}", cwd="ns3").out.strip()
    settime(
        f"-g HIDDEN -k UNRETENTIVE {TremN} -r UNRETENTIVE {TremN} -z HIDDEN now-2295h -d HIDDEN {TremN} {csk1_name}",
        cwd="ns3",
    )
    settime(
        f"-g OMNIPRESENT -k OMNIPRESENT {TactN1} -r OMNIPRESENT {TactN1} -z OMNIPRESENT now-2295h -d OMNIPRESENT {TremN} {csk2_name}",
        cwd="ns3",
    )
    # Set key rollover relationship.
    set_key_relationship(csk1_name, csk2_name)
    # Signing.
    render_and_sign_zone(zonename, [csk1_name, csk2_name], extra_options=f"-z -G {cds}")

    return zones


def configure_enable_dnssec(tld: str, policy: str) -> List[Zone]:
    # The zones at enable-dnssec.$tld represent the various steps of the
    # initial signing of a zone.
    zones = []
    zone = f"enable-dnssec.{tld}"
    keygen = EnvCmd("KEYGEN", f"-k {policy} -l kasp.conf")
    settime = EnvCmd("SETTIME", "-s")

    # Step 1:
    # This is an unsigned zone and named should perform the initial steps of
    # introducing the DNSSEC records in the right order.
    zonename = f"step1.{zone}"
    zones.append(Zone(zonename, f"{zonename}.db", Nameserver("ns3", "10.53.0.3")))
    isctest.log.info(f"setup {zonename}")
    render_and_sign_zone(zonename, [], signing=False)

    # Step 2:
    # The DNSKEY has been published long enough to become OMNIPRESENT.
    zonename = f"step2.{zone}"
    zones.append(Zone(zonename, f"{zonename}.db", Nameserver("ns3", "10.53.0.3")))
    isctest.log.info(f"setup {zonename}")
    # DNSKEY TTL:             300 seconds
    # zone-propagation-delay: 5 minutes (300 seconds)
    # publish-safety:         5 minutes (300 seconds)
    # Total:                  900 seconds
    TpubN = "now-900s"
    keytimes = f"-P {TpubN} -A {TpubN}"
    # Key generation.
    csk_name = keygen(f"{keytimes} {zonename}", cwd="ns3").out.strip()
    settime(
        f"-g OMNIPRESENT -k RUMOURED {TpubN} -r RUMOURED {TpubN} -z RUMOURED {TpubN} -d HIDDEN {TpubN} {csk_name}",
        cwd="ns3",
    )
    # Signing.
    render_and_sign_zone(zonename, [csk_name], extra_options="-z")

    # Step 3:
    # The zone signatures have been published long enough to become OMNIPRESENT.
    zonename = f"step3.{zone}"
    zones.append(Zone(zonename, f"{zonename}.db", Nameserver("ns3", "10.53.0.3")))
    isctest.log.info(f"setup {zonename}")
    # Passed time since publication:
    # max-zone-ttl:           12 hours (43200 seconds)
    # zone-propagation-delay: 5 minutes (300 seconds)
    # We can submit the DS now.
    TpubN = "now-43500s"
    keytimes = f"-P {TpubN} -A {TpubN}"
    # Key generation.
    csk_name = keygen(f"{keytimes} {zonename}", cwd="ns3").out.strip()
    settime(
        f"-g OMNIPRESENT -k OMNIPRESENT {TpubN} -r OMNIPRESENT {TpubN} -z RUMOURED {TpubN} -d HIDDEN {TpubN} {csk_name}",
        cwd="ns3",
    )
    # Signing.
    render_and_sign_zone(zonename, [csk_name], extra_options="-z")

    # Step 4:
    # The DS has been submitted long enough ago to become OMNIPRESENT.
    zonename = f"step4.{zone}"
    zones.append(Zone(zonename, f"{zonename}.db", Nameserver("ns3", "10.53.0.3")))
    isctest.log.info(f"setup {zonename}")
    # DS TTL:                    2 hour (7200 seconds)
    # parent-propagation-delay:  1 hour (3600 seconds)
    # Total aditional time:      10800 seconds
    # 43500 + 10800 = 54300
    TpubN = "now-54300s"
    TsbmN = "now-10800s"
    keytimes = f"-P {TpubN} -A {TpubN} -P sync {TsbmN}"
    # Key generation.
    csk_name = keygen(f"{keytimes} {zonename}", cwd="ns3").out.strip()
    settime(
        f"-g OMNIPRESENT -k OMNIPRESENT {TpubN} -r OMNIPRESENT {TpubN} -z OMNIPRESENT {TsbmN} -d RUMOURED {TpubN} -P ds {TsbmN} {csk_name}",
        cwd="ns3",
    )
    # Signing.
    render_and_sign_zone(zonename, [csk_name], extra_options="-z")

    return zones


def configure_going_insecure(tld: str, reconfig: bool = False) -> List[Zone]:
    zones = []
    keygen = EnvCmd("KEYGEN", "-a ECDSA256 -L 7200")
    settime = EnvCmd("SETTIME", "-s")

    # The child zones (step1, step2) beneath these zones represent the various
    # steps of unsigning a zone.
    for zone in [f"going-insecure.{tld}", f"going-insecure-dynamic.{tld}"]:
        # Set up a zone with dnssec-policy that is going insecure.

        # Step 1:
        zonename = f"step1.{zone}"
        zones.append(Zone(zonename, f"{zonename}.db", Nameserver("ns3", "10.53.0.3")))
        isctest.log.info(f"setup {zonename}")
        # Timing metadata.
        TpubN = "now-10d"
        TsbmN = "now-12955mi"
        keytimes = f"-P {TpubN} -A {TpubN}"
        cdstimes = f"-P sync {TsbmN}"
        # Key generation.
        ksk_name = keygen(
            f"-f KSK {keytimes} {cdstimes} {zonename}", cwd="ns3"
        ).out.strip()
        zsk_name = keygen(f"{keytimes} {zonename}", cwd="ns3").out.strip()
        settime(
            f"-g OMNIPRESENT -k OMNIPRESENT {TpubN} -r OMNIPRESENT {TpubN} -d OMNIPRESENT {TpubN} {ksk_name}",
            cwd="ns3",
        )
        settime(
            f"-g OMNIPRESENT -k OMNIPRESENT {TpubN} -z OMNIPRESENT {TpubN} {zsk_name}",
            cwd="ns3",
        )
        # Signing.
        render_and_sign_zone(zonename, [ksk_name, zsk_name])

        if reconfig:
            # Step 2:
            zonename = f"step2.{zone}"
            zones.append(
                Zone(zonename, f"{zonename}.db", Nameserver("ns3", "10.53.0.3"))
            )
            isctest.log.info(f"setup {zonename}")
            # The DS was withdrawn from the parent zone 26 hours ago.
            TremN = "now-26h"
            keytimes = f"-P {TpubN} -A {TpubN} -I {TremN} -D now"
            cdstimes = f"-P sync {TsbmN} -D sync {TremN}"
            # Key generation.
            ksk_name = keygen(
                f"-f KSK {keytimes} {cdstimes} {zonename}", cwd="ns3"
            ).out.strip()
            zsk_name = keygen(f"{keytimes} {zonename}", cwd="ns3").out.strip()
            settime(
                f"-g HIDDEN -k OMNIPRESENT {TpubN} -r OMNIPRESENT {TpubN} -d UNRETENTIVE {TremN} -D ds {TremN} {ksk_name}",
                cwd="ns3",
            )
            settime(
                f"-g HIDDEN -k OMNIPRESENT {TpubN} -z OMNIPRESENT {TpubN} {zsk_name}",
                cwd="ns3",
            )
            # Fake lifetime of old algorithm keys.
            fake_lifetime(ksk_name, 0)
            fake_lifetime(zsk_name, 5184000)
            # Signing.
            render_and_sign_zone(zonename, [ksk_name, zsk_name], extra_options="-P")

    return zones


def configure_straight2none(tld: str) -> List[Zone]:
    # These zones are going straight to "none" policy. This is undefined behavior.
    zones = []
    keygen = EnvCmd("KEYGEN", "-k default")
    settime = EnvCmd("SETTIME", "-s")

    TpubN = "now-10d"
    TsbmN = "now-12955mi"
    keytimes = f"-P {TpubN} -A {TpubN} -P sync {TsbmN}"

    zonename = f"going-straight-to-none.{tld}"
    zones.append(Zone(zonename, f"{zonename}.db", Nameserver("ns3", "10.53.0.3")))
    isctest.log.info(f"setup {zonename}")
    # Key generation.
    csk_name = keygen(f"-f KSK {keytimes} {zonename}", cwd="ns3").out.strip()
    settime(
        f"-g OMNIPRESENT -k OMNIPRESENT {TpubN} -r OMNIPRESENT {TpubN} -z OMNIPRESENT {TpubN} -d OMNIPRESENT {TpubN} {csk_name}",
        cwd="ns3",
    )
    # Signing.
    render_and_sign_zone(zonename, [csk_name], extra_options="-z")

    zonename = f"going-straight-to-none-dynamic.{tld}"
    zones.append(
        Zone(zonename, f"{zonename}.db.signed", Nameserver("ns3", "10.53.0.3"))
    )
    isctest.log.info(f"setup {zonename}")
    # Key generation.
    csk_name = keygen(f"-f KSK {keytimes} {zonename}", cwd="ns3").out.strip()
    settime(
        f"-g OMNIPRESENT -k OMNIPRESENT {TpubN} -r OMNIPRESENT {TpubN} -z OMNIPRESENT {TpubN} -d OMNIPRESENT {TpubN} {csk_name}",
        cwd="ns3",
    )
    # Signing.
    render_and_sign_zone(zonename, [csk_name], extra_options="-z -O full")

    return zones


def configure_ksk_doubleksk(tld: str) -> List[Zone]:
    # The zones at ksk-doubleksk.$tld represent the various steps of a KSK
    # Double-KSK rollover.
    zones = []
    zone = f"ksk-doubleksk.{tld}"
    cds = "cds:sha-256"
    keygen = EnvCmd("KEYGEN", "-a ECDSAP256SHA256 -L 7200")
    settime = EnvCmd("SETTIME", "-s")

    # Step 1:
    # Introduce the first key. This will immediately be active.
    zonename = f"step1.{zone}"
    zones.append(Zone(zonename, f"{zonename}.db", Nameserver("ns3", "10.53.0.3")))
    isctest.log.info(f"setup {zonename}")
    # Timing metadata.
    TactN = "now-7d"
    keytimes = f"-P {TactN} -A {TactN}"
    # Key generation.
    ksk_name = keygen(f"-f KSK {keytimes} {zonename}", cwd="ns3").out.strip()
    zsk_name = keygen(f"{keytimes} {zonename}", cwd="ns3").out.strip()
    settime(
        f"-g OMNIPRESENT -k OMNIPRESENT {TactN} -r OMNIPRESENT {TactN} -d OMNIPRESENT {TactN} {ksk_name}",
        cwd="ns3",
    )
    settime(
        f"-g OMNIPRESENT -k OMNIPRESENT {TactN} -z OMNIPRESENT {TactN} {zsk_name}",
        cwd="ns3",
    )
    # Signing.
    render_and_sign_zone(zonename, [ksk_name, zsk_name], extra_options=f"-G {cds}")

    # Step 2:
    # It is time to introduce the new KSK.
    zonename = f"step2.{zone}"
    zones.append(Zone(zonename, f"{zonename}.db", Nameserver("ns3", "10.53.0.3")))
    isctest.log.info(f"setup {zonename}")
    # Lksk:           60d
    # Dreg:           n/a
    # DprpC:          1h
    # TTLds:          1d
    # TTLkey:         2h
    # publish-safety: 1d
    # retire-safety:  2d
    #
    # According to RFC 7583:
    # Tpub(N+1) <= Tact(N) + Lksk - Dreg - IpubC
    # IpubC = DprpC + TTLkey (+publish-safety)
    #
    # IpubC   = 27h
    # Tact(N) = now - Lksk + Dreg + IpubC = now - 60d + 27h
    #         = now - 1440h + 27h = now - 1413h
    TactN = "now-1413h"
    keytimes = f"-P {TactN} -A {TactN}"
    # Key generation.
    ksk_name = keygen(f"-f KSK {keytimes} {zonename}", cwd="ns3").out.strip()
    zsk_name = keygen(f"{keytimes} {zonename}", cwd="ns3").out.strip()
    settime(
        f"-g OMNIPRESENT -k OMNIPRESENT {TactN} -r OMNIPRESENT {TactN} -d OMNIPRESENT {TactN} {ksk_name}",
        cwd="ns3",
    )
    settime(
        f"-g OMNIPRESENT -k OMNIPRESENT {TactN} -z OMNIPRESENT {TactN} {zsk_name}",
        cwd="ns3",
    )
    # Signing.
    render_and_sign_zone(zonename, [ksk_name, zsk_name], extra_options=f"-G {cds}")

    # Step 3:
    # It is time to submit the DS.
    zonename = f"step3.{zone}"
    zones.append(Zone(zonename, f"{zonename}.db", Nameserver("ns3", "10.53.0.3")))
    isctest.log.info(f"setup {zonename}")
    # According to RFC 7583:
    # Iret = DprpP + TTLds (+retire-safety)
    #
    # Iret       = 50h
    # Tpub(N)    = now - Lksk = now - 60d = now - 60d
    # Tact(N)    = now - 1413h
    # Tret(N)    = now
    # Trem(N)    = now + Iret = now + 50h
    # Tpub(N+1)  = now - IpubC = now - 27h
    # Tact(N+1)  = now
    # Tret(N+1)  = now + Lksk = now + 60d
    # Trem(N+1)  = now + Lksk + Iret = now + 60d + 50h
    #            = now + 1440h + 50h = 1490h
    TpubN = "now-60d"
    TactN = "now-1413h"
    TretN = "now"
    TremN = "now+50h"
    TpubN1 = "now-27h"
    TactN1 = "now"
    TretN1 = "now+60d"
    TremN1 = "now+1490h"
    ksktimes = (
        f"-P {TpubN} -A {TpubN} -P sync {TactN} -I {TretN} -D {TremN} -D sync {TactN1}"
    )
    newtimes = f"-P {TpubN1} -A {TactN1} -P sync {TactN1} -I {TretN1} -D {TremN1}"
    zsktimes = f"-P {TpubN}  -A {TpubN}"
    # Key generation.
    ksk1_name = keygen(f"-f KSK {ksktimes} {zonename}", cwd="ns3").out.strip()
    ksk2_name = keygen(f"-f KSK {newtimes} {zonename}", cwd="ns3").out.strip()
    zsk_name = keygen(f"{zsktimes} {zonename}", cwd="ns3").out.strip()
    settime(
        f"-g HIDDEN -k OMNIPRESENT {TactN} -r OMNIPRESENT {TactN} -d OMNIPRESENT {TactN} {ksk1_name}",
        cwd="ns3",
    )
    settime(
        f"-g OMNIPRESENT -k RUMOURED {TpubN1} -r RUMOURED {TpubN1} -d HIDDEN {TpubN1} {ksk2_name}",
        cwd="ns3",
    )
    settime(
        f"-g OMNIPRESENT -k OMNIPRESENT {TpubN} -z OMNIPRESENT {TpubN} {zsk_name}",
        cwd="ns3",
    )
    # Set key rollover relationship.
    set_key_relationship(ksk1_name, ksk2_name)
    # Signing.
    render_and_sign_zone(
        zonename, [ksk1_name, ksk2_name, zsk_name], extra_options=f"-G {cds}"
    )

    # Step 4:
    # The DS should be swapped now.
    zonename = f"step4.{zone}"
    zones.append(Zone(zonename, f"{zonename}.db", Nameserver("ns3", "10.53.0.3")))
    isctest.log.info(f"setup {zonename}")
    # Tpub(N)    = now - Lksk - Iret = now - 60d - 50h
    #            = now - 1440h - 50h = now - 1490h
    # Tact(N)    = now - 1490h + 27h = now - 1463h
    # Tret(N)    = now - Iret = now - 50h
    # Trem(N)    = now
    # Tpub(N+1)  = now - Iret - IpubC = now - 50h - 27h
    #            = now - 77h
    # Tact(N+1)  = Tret(N)
    # Tret(N+1)  = now + Lksk - Iret = now + 60d - 50h = now + 1390h
    # Trem(N+1)  = now + Lksk = now + 60d
    TpubN = "now-1490h"
    TactN = "now-1463h"
    TretN = "now-50h"
    TremN = "now"
    TpubN1 = "now-77h"
    TactN1 = TretN
    TretN1 = "now+1390h"
    TremN1 = "now+60d"
    ksktimes = (
        f"-P {TpubN} -A {TpubN} -P sync {TactN} -I {TretN} -D {TremN} -D sync {TactN1}"
    )
    newtimes = f"-P {TpubN1} -A {TactN1} -P sync {TactN1} -I {TretN1} -D {TremN1}"
    zsktimes = f"-P {TpubN} -A {TpubN}"
    # Key generation.
    ksk1_name = keygen(f"-f KSK {ksktimes} {zonename}", cwd="ns3").out.strip()
    ksk2_name = keygen(f"-f KSK {newtimes} {zonename}", cwd="ns3").out.strip()
    zsk_name = keygen(f"{zsktimes} {zonename}", cwd="ns3").out.strip()
    settime(
        f"-g HIDDEN -k OMNIPRESENT {TactN} -r OMNIPRESENT {TactN} -d UNRETENTIVE {TretN} -D ds {TretN} {ksk1_name}",
        cwd="ns3",
    )
    settime(
        f"-g OMNIPRESENT -k OMNIPRESENT {TactN1} -r OMNIPRESENT {TactN1} -d RUMOURED {TactN1} -P ds {TactN1} {ksk2_name}",
        cwd="ns3",
    )
    settime(
        f"-g OMNIPRESENT -k OMNIPRESENT {TactN} -z OMNIPRESENT {TactN} {zsk_name}",
        cwd="ns3",
    )
    # Set key rollover relationship.
    set_key_relationship(ksk1_name, ksk2_name)
    # Signing.
    render_and_sign_zone(
        zonename, [ksk1_name, ksk2_name, zsk_name], extra_options=f"-G {cds}"
    )

    # Step 5:
    # The predecessor DNSKEY is removed long enough that is has become HIDDEN.
    zonename = f"step5.{zone}"
    zones.append(Zone(zonename, f"{zonename}.db", Nameserver("ns3", "10.53.0.3")))
    isctest.log.info(f"setup {zonename}")
    # Subtract DNSKEY TTL + zone-propagation-delay from all the times (3h).
    # Tpub(N)    = now - 1490h - 3h = now - 1493h
    # Tact(N)    = now - 1463h - 3h = now - 1466h
    # Tret(N)    = now - 50h - 3h = now - 53h
    # Trem(N)    = now - 3h
    # Tpub(N+1)  = now - 77h - 3h = now - 80h
    # Tact(N+1)  = Tret(N)
    # Tret(N+1)  = now + 1390h - 3h = now + 1387h
    # Trem(N+1)  = now + 60d - 3h = now + 1441h
    TpubN = "now-1493h"
    TactN = "now-1466h"
    TretN = "now-53h"
    TremN = "now-3h"
    TpubN1 = "now-80h"
    TactN1 = TretN
    TretN1 = "now+1387h"
    TremN1 = "now+1441h"
    ksktimes = (
        f"-P {TpubN} -A {TpubN} -P sync {TactN} -I {TretN} -D {TremN} -D sync {TactN1}"
    )
    newtimes = f"-P {TpubN1} -A {TactN1} -P sync {TactN1} -I {TretN1} -D {TremN1}"
    zsktimes = f"-P {TpubN} -A {TpubN}"
    # Key generation.
    ksk1_name = keygen(f"-f KSK {ksktimes} {zonename}", cwd="ns3").out.strip()
    ksk2_name = keygen(f"-f KSK {newtimes} {zonename}", cwd="ns3").out.strip()
    zsk_name = keygen(f"{zsktimes} {zonename}", cwd="ns3").out.strip()
    settime(
        f"-g HIDDEN -k UNRETENTIVE {TretN} -r UNRETENTIVE {TretN} -d HIDDEN {TretN} {ksk1_name}",
        cwd="ns3",
    )
    settime(
        f"-g OMNIPRESENT -k OMNIPRESENT {TactN1} -r OMNIPRESENT {TactN1} -d OMNIPRESENT {TactN1} {ksk2_name}",
        cwd="ns3",
    )
    settime(
        f"-g OMNIPRESENT -k OMNIPRESENT {TactN} -z OMNIPRESENT {TactN} {zsk_name}",
        cwd="ns3",
    )
    # Set key rollover relationship.
    set_key_relationship(ksk1_name, ksk2_name)
    # Signing.
    render_and_sign_zone(
        zonename, [ksk1_name, ksk2_name, zsk_name], extra_options=f"-G {cds}"
    )

    # Step 6:
    # The predecessor DNSKEY can be purged.
    zonename = f"step6.{zone}"
    zones.append(Zone(zonename, f"{zonename}.db", Nameserver("ns3", "10.53.0.3")))
    isctest.log.info(f"setup {zonename}")
    # Subtract purge-keys interval from all the times (1h).
    TpubN = "now-1494h"
    TactN = "now-1467h"
    TretN = "now-54h"
    TremN = "now-4h"
    TpubN1 = "now-81h"
    TactN1 = TretN
    TretN1 = "now+1386h"
    TremN1 = "now+1440h"
    ksktimes = (
        f"-P {TpubN} -A {TpubN} -P sync {TactN} -I {TretN} -D {TremN} -D sync {TactN1}"
    )
    newtimes = f"-P {TpubN1} -A {TactN1} -P sync {TactN1} -I {TretN1} -D {TremN1}"
    zsktimes = f"-P {TpubN} -A {TpubN}"
    # Key generation.
    ksk1_name = keygen(f"-f KSK {ksktimes} {zonename}", cwd="ns3").out.strip()
    ksk2_name = keygen(f"-f KSK {newtimes} {zonename}", cwd="ns3").out.strip()
    zsk_name = keygen(f"{zsktimes} {zonename}", cwd="ns3").out.strip()
    settime(
        f"-g HIDDEN -k HIDDEN {TretN} -r HIDDEN {TretN} -d HIDDEN {TretN} {ksk1_name}",
        cwd="ns3",
    )
    settime(
        f"-g OMNIPRESENT -k OMNIPRESENT {TactN1} -r OMNIPRESENT {TactN1} -d OMNIPRESENT {TactN1} {ksk2_name}",
        cwd="ns3",
    )
    settime(
        f"-g OMNIPRESENT -k OMNIPRESENT {TactN} -z OMNIPRESENT {TactN} {zsk_name}",
        cwd="ns3",
    )
    # Set key rollover relationship.
    set_key_relationship(ksk1_name, ksk2_name)
    # Signing.
    render_and_sign_zone(
        zonename, [ksk1_name, ksk2_name, zsk_name], extra_options=f"-G {cds}"
    )

    return zones


def configure_ksk_3crowd(tld: str) -> List[Zone]:
    # Test #2375, the "three is a crowd" bug, where a new key is introduced but the
    # previous rollover has not finished yet. In other words, we have a key KEY2
    # that is the successor of key KEY1, and we introduce a new key KEY3 that is
    # the successor of key KEY2:
    #
    #     KEY1 < KEY2 < KEY3.
    #
    # The expected behavior is that all three keys remain in the zone, and not
    # the bug behavior where KEY2 is removed and immediately replaced with KEY3.
    #
    zones = []
    cds = "cds:sha-256"
    keygen = EnvCmd("KEYGEN", "-a ECDSAP256SHA256 -L 7200")
    settime = EnvCmd("SETTIME", "-s")

    # Set up a zone that has a KSK (KEY1) and have the successor key (KEY2)
    # published as well.
    zonename = f"three-is-a-crowd.{tld}"
    zones.append(Zone(zonename, f"{zonename}.db", Nameserver("ns3", "10.53.0.3")))
    isctest.log.info(f"setup {zonename}")
    # These times are the same as step3.ksk-doubleksk.autosign.
    TpubN = "now-60d"
    TactN = "now-1413h"
    TretN = "now"
    TremN = "now+50h"
    TpubN1 = "now-27h"
    TactN1 = TretN
    TretN1 = "now+60d"
    TremN1 = "now+1490h"
    ksktimes = (
        f"-P {TpubN} -A {TpubN} -P sync {TactN} -I {TretN} -D {TremN} -D sync {TactN1}"
    )
    newtimes = f"-P {TpubN1} -A {TactN1} -P sync {TactN1} -I {TretN1} -D {TremN1}"
    zsktimes = f"-P {TpubN}  -A {TpubN}"
    # Key generation.
    ksk1_name = keygen(f"-f KSK {ksktimes} {zonename}", cwd="ns3").out.strip()
    ksk2_name = keygen(f"-f KSK {newtimes} {zonename}", cwd="ns3").out.strip()
    zsk_name = keygen(f"{zsktimes} {zonename}", cwd="ns3").out.strip()
    settime(
        f"-g HIDDEN -k OMNIPRESENT {TactN} -r OMNIPRESENT {TactN} -d OMNIPRESENT {TactN} {ksk1_name}",
        cwd="ns3",
    )
    settime(
        f"-g OMNIPRESENT -k RUMOURED {TpubN1} -r RUMOURED {TpubN1} -d HIDDEN {TpubN1} {ksk2_name}",
        cwd="ns3",
    )
    settime(
        f"-g OMNIPRESENT -k OMNIPRESENT {TpubN} -z OMNIPRESENT {TpubN} {zsk_name}",
        cwd="ns3",
    )
    # Set key rollover relationship.
    set_key_relationship(ksk1_name, ksk2_name)
    # Signing.
    render_and_sign_zone(
        zonename, [ksk1_name, ksk2_name, zsk_name], extra_options=f"-G {cds}"
    )

    return zones


def configure_zsk_prepub(tld: str) -> List[Zone]:
    # The zones at zsk-prepub.$tld represent the various steps of a ZSK
    # Pre-Publication rollover.
    zones = []
    zone = f"zsk-prepub.{tld}"
    keygen = EnvCmd("KEYGEN", "-a ECDSAP256SHA256 -L 3600")
    settime = EnvCmd("SETTIME", "-s")

    # Step 1:
    # Introduce the first key. This will immediately be active.
    zonename = f"step1.{zone}"
    zones.append(Zone(zonename, f"{zonename}.db", Nameserver("ns3", "10.53.0.3")))
    isctest.log.info(f"setup {zonename}")
    # Timing metadata.
    TactN = "now-7d"
    keytimes = f"-P {TactN} -A {TactN}"
    # Key generation.
    ksk_name = keygen(f"-f KSK {keytimes} {zonename}", cwd="ns3").out.strip()
    zsk_name = keygen(f"{keytimes} {zonename}", cwd="ns3").out.strip()
    settime(
        f"-g OMNIPRESENT -k OMNIPRESENT {TactN} -r OMNIPRESENT {TactN} -d OMNIPRESENT {TactN} {ksk_name}",
        cwd="ns3",
    )
    settime(
        f"-g OMNIPRESENT -k OMNIPRESENT {TactN} -z OMNIPRESENT {TactN} {zsk_name}",
        cwd="ns3",
    )
    # Signing.
    render_and_sign_zone(zonename, [ksk_name, zsk_name])

    # Step 2:
    # It is time to pre-publish the successor ZSK.
    zonename = f"step2.{zone}"
    zones.append(Zone(zonename, f"{zonename}.db", Nameserver("ns3", "10.53.0.3")))
    isctest.log.info(f"setup {zonename}")
    # According to RFC 7583:
    # Tact(N) = now + Ipub - Lzsk = now + 26h - 30d
    #         = now + 26h - 30d = now − 694h
    TactN = "now-694h"
    keytimes = f"-P {TactN} -A {TactN}"
    # Key generation.
    ksk_name = keygen(f"-f KSK {keytimes} {zonename}", cwd="ns3").out.strip()
    zsk_name = keygen(f"{keytimes} {zonename}", cwd="ns3").out.strip()
    settime(
        f"-g OMNIPRESENT -k OMNIPRESENT {TactN} -r OMNIPRESENT {TactN} -d OMNIPRESENT {TactN} {ksk_name}",
        cwd="ns3",
    )
    settime(
        f"-g OMNIPRESENT -k OMNIPRESENT {TactN} -z OMNIPRESENT {TactN} {zsk_name}",
        cwd="ns3",
    )
    # Signing.
    render_and_sign_zone(zonename, [ksk_name, zsk_name])

    # Step 3:
    # After the publication interval has passed the DNSKEY of the successor ZSK
    # is OMNIPRESENT and the zone can thus be signed with the successor ZSK.
    zonename = f"step3.{zone}"
    zones.append(Zone(zonename, f"{zonename}.db", Nameserver("ns3", "10.53.0.3")))
    isctest.log.info(f"setup {zonename}")
    # According to RFC 7583:
    # Tpub(N+1) <= Tact(N) + Lzsk - Ipub
    # Tact(N+1) = Tact(N) + Lzsk
    #
    # Tact(N)   = now - Lzsk = now - 30d
    # Tpub(N+1) = now - Ipub = now - 26h
    # Tact(N+1) = now
    # Tret(N) = now
    # Trem(N) = now + Iret = now + Dsign + Dprp + TTLsig + retire-safety = 8d1h = now + 241h
    TactN = "now-30d"
    TpubN1 = "now-26h"
    TactN1 = "now"
    TremN = "now+241h"
    keytimes = f"-P {TactN} -A {TactN}"
    oldtimes = f"-P {TactN} -A {TactN} -I {TactN1} -D {TremN}"
    newtimes = f"-P {TpubN1} -A {TactN1}"
    # Key generation.
    ksk_name = keygen(f"-f KSK {keytimes} {zonename}", cwd="ns3").out.strip()
    zsk1_name = keygen(f"{oldtimes} {zonename}", cwd="ns3").out.strip()
    zsk2_name = keygen(f"{newtimes} {zonename}", cwd="ns3").out.strip()
    settime(
        f"-g OMNIPRESENT -k OMNIPRESENT {TactN} -r OMNIPRESENT {TactN} -d OMNIPRESENT {TactN} {ksk_name}",
        cwd="ns3",
    )
    settime(
        f"-g HIDDEN -k OMNIPRESENT {TactN} -z OMNIPRESENT {TactN} {zsk1_name}",
        cwd="ns3",
    )
    settime(
        f"-g OMNIPRESENT -k RUMOURED {TpubN1} -z HIDDEN {TpubN1} {zsk2_name}", cwd="ns3"
    )
    # Set key rollover relationship.
    set_key_relationship(zsk1_name, zsk2_name)
    # Signing.
    render_and_sign_zone(zonename, [ksk_name, zsk1_name, zsk2_name])

    # Step 4:
    # After the retire interval has passed the predecessor DNSKEY can be
    # removed from the zone.
    zonename = f"step4.{zone}"
    zones.append(Zone(zonename, f"{zonename}.db", Nameserver("ns3", "10.53.0.3")))
    isctest.log.info(f"setup {zonename}")
    # Lzsk:          30d
    # Ipub:          26h
    # Dsgn:          1w
    # Dprp:          1h
    # TTLsig:        1d
    # retire-safety: 2d
    #
    # According to RFC 7583:
    # Iret      = Dsgn + Dprp + TTLsig (+retire-safety)
    # Iret      = 1w + 1h + 1d + 2d = 10d1h = 241h
    #
    # Tact(N)   = now - Iret - Lzsk
    #           = now - 241h - 30d = now - 241h - 720h
    #           = now - 961h
    # Tpub(N+1) = now - Iret - Ipub
    #           = now - 241h - 26h
    #           = now - 267h
    # Tact(N+1) = now - Iret = now - 241h
    TactN = "now-961h"
    TpubN1 = "now-267h"
    TactN1 = "now-241h"
    TremN = "now"
    keytimes = f"-P {TactN} -A {TactN}"
    oldtimes = f"-P {TactN} -A {TactN} -I {TactN1} -D {TremN}"
    newtimes = f"-P {TpubN1} -A {TactN1}"
    # Key generation.
    ksk_name = keygen(f"-f KSK {keytimes} {zonename}", cwd="ns3").out.strip()
    zsk1_name = keygen(f"{oldtimes} {zonename}", cwd="ns3").out.strip()
    zsk2_name = keygen(f"{newtimes} {zonename}", cwd="ns3").out.strip()
    settime(
        f"-g OMNIPRESENT -k OMNIPRESENT {TactN} -r OMNIPRESENT {TactN} -d OMNIPRESENT {TactN} {ksk_name}",
        cwd="ns3",
    )
    settime(
        f"-g HIDDEN -k OMNIPRESENT {TactN} -z UNRETENTIVE {TactN1} {zsk1_name}",
        cwd="ns3",
    )
    settime(
        f"-g OMNIPRESENT -k OMNIPRESENT {TactN1} -z RUMOURED {TactN1} {zsk2_name}",
        cwd="ns3",
    )
    # Set key rollover relationship.
    set_key_relationship(zsk1_name, zsk2_name)
    # Signing.
    render_and_sign_zone(zonename, [ksk_name, zsk1_name, zsk2_name])

    # Step 5:
    # The predecessor DNSKEY is removed long enough that is has become HIDDEN.
    zonename = f"step5.{zone}"
    zones.append(Zone(zonename, f"{zonename}.db", Nameserver("ns3", "10.53.0.3")))
    isctest.log.info(f"setup {zonename}")
    # Subtract DNSKEY TTL + zone-propagation-delay from all the times (2h).
    # Tact(N)   = now - 961h - 2h = now - 963h
    # Tpub(N+1) = now - 267h - 2h = now - 269h
    # Tact(N+1) = now - 241h - 2h = now - 243h
    # Trem(N)   = Tact(N+1) + Iret = now -2h
    TactN = "now-963h"
    TremN = "now-2h"
    TpubN1 = "now-269h"
    TactN1 = "now-243h"
    TremN = "now-2h"
    keytimes = f"-P {TactN} -A {TactN}"
    oldtimes = f"-P {TactN} -A {TactN} -I {TactN1} -D {TremN}"
    newtimes = f"-P {TpubN1} -A {TactN1}"
    # Key generation.
    ksk_name = keygen(f"-f KSK {keytimes} {zonename}", cwd="ns3").out.strip()
    zsk1_name = keygen(f"{oldtimes} {zonename}", cwd="ns3").out.strip()
    zsk2_name = keygen(f"{newtimes} {zonename}", cwd="ns3").out.strip()
    settime(
        f"-g OMNIPRESENT -k OMNIPRESENT {TactN} -r OMNIPRESENT {TactN} -d OMNIPRESENT {TactN} {ksk_name}",
        cwd="ns3",
    )
    settime(
        f"-g HIDDEN -k UNRETENTIVE {TremN} -z HIDDEN {TremN} {zsk1_name}", cwd="ns3"
    )
    settime(
        f"-g OMNIPRESENT -k OMNIPRESENT {TactN1} -z OMNIPRESENT {TremN} {zsk2_name}",
        cwd="ns3",
    )
    # Set key rollover relationship.
    set_key_relationship(zsk1_name, zsk2_name)
    # Signing.
    render_and_sign_zone(zonename, [ksk_name, zsk1_name, zsk2_name])

    # Step 6:
    # The predecessor DNSKEY can be purged.
    zonename = f"step6.{zone}"
    zones.append(Zone(zonename, f"{zonename}.db", Nameserver("ns3", "10.53.0.3")))
    isctest.log.info(f"setup {zonename}")
    # Subtract purge-keys interval from all the times (1h).
    TactN = "now-964h"
    TremN = "now-3h"
    TpubN1 = "now-270h"
    TactN1 = "now-244h"
    TremN = "now-3h"
    keytimes = f"-P {TactN} -A {TactN}"
    oldtimes = f"-P {TactN} -A {TactN} -I {TactN1} -D {TremN}"
    newtimes = f"-P {TpubN1} -A {TactN1}"
    # Key generation.
    ksk_name = keygen(f"-f KSK {keytimes} {zonename}", cwd="ns3").out.strip()
    zsk1_name = keygen(f"{oldtimes} {zonename}", cwd="ns3").out.strip()
    zsk2_name = keygen(f"{newtimes} {zonename}", cwd="ns3").out.strip()
    settime(
        f"-g OMNIPRESENT -k OMNIPRESENT {TactN} -r OMNIPRESENT {TactN} -d OMNIPRESENT {TactN} {ksk_name}",
        cwd="ns3",
    )
    settime(f"-g HIDDEN -k HIDDEN {TremN} -z HIDDEN {TremN} {zsk1_name}", cwd="ns3")
    settime(
        f"-g OMNIPRESENT -k OMNIPRESENT {TactN1} -z OMNIPRESENT {TremN} {zsk2_name}",
        cwd="ns3",
    )
    # Set key rollover relationship.
    set_key_relationship(zsk1_name, zsk2_name)
    # Signing.
    render_and_sign_zone(zonename, [ksk_name, zsk1_name, zsk2_name])

    return zones



# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

import dns.message
import pytest

import isctest

pytestmark = pytest.mark.extra_artifacts(
    [
        "ns2/example.bk",
        "ns3/example.bk",
    ]
)


@pytest.mark.parametrize("ns", [2, 3])
def test_dialup_zone_transfer(named_port, servers, ns):
    msg = dns.message.make_query("example.", "SOA")
    # Drop the RD flag from the query
    msg.flags &= ~dns.flags.RD
    ns1response = isctest.query.tcp(msg, "10.53.0.1")
    with servers[f"ns{ns}"].watch_log_from_start(timeout=90) as watcher:
        watcher.wait_for_line(
            f"transfer of 'example/IN' from 10.53.0.{ns-1}#{named_port}: Transfer status: success",
        )
    response = isctest.query.tcp(msg, f"10.53.0.{ns}")
    if response.rcode() != dns.rcode.SERVFAIL:
        assert response.answer == ns1response.answer
        assert response.authority == ns1response.authority



"""
Copyright (C) Internet Systems Consortium, Inc. ("ISC")

SPDX-License-Identifier: MPL-2.0

This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0.  If a copy of the MPL was not distributed with this
file, you can obtain one at https://mozilla.org/MPL/2.0/.

See the COPYRIGHT file distributed with this work for additional
information regarding copyright ownership.
"""

from typing import AsyncGenerator

import dns.rdatatype
import dns.rrset

from isctest.asyncserver import (
    ControllableAsyncDnsServer,
    DnsResponseSend,
    DomainHandler,
    QueryContext,
    ResponseAction,
    ToggleResponsesCommand,
)


class ExtraAnswersHandler(DomainHandler):
    """
    Answer from zone data, inserting extra RRsets into responses to A queries.
    """

    domains = ["attackSecureDomain.net3."]

    async def get_responses(
        self, qctx: QueryContext
    ) -> AsyncGenerator[ResponseAction, None]:
        if qctx.qtype == dns.rdatatype.A:
            ns_rrset = dns.rrset.from_text(
                "net3.", 300, qctx.qclass, dns.rdatatype.NS, "local.net3."
            )
            qctx.response.answer.append(ns_rrset)
            a_rrset = dns.rrset.from_text(
                "local.net3.", 300, qctx.qclass, dns.rdatatype.A, "10.53.0.11"
            )
            qctx.response.additional.append(a_rrset)

        yield DnsResponseSend(qctx.response)


def main() -> None:
    server = ControllableAsyncDnsServer()
    server.install_control_command(ToggleResponsesCommand())
    server.install_response_handler(ExtraAnswersHandler())
    server.run()


if __name__ == "__main__":
    main()



# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

import pytest

import isctest.mark

pytestmark = [
    isctest.mark.dnsrps_enabled,
    pytest.mark.extra_artifacts(
        [
            "dig.out*",
            "dnsrps.cache",
            "dnsrps.conf",
            "dnsrps.zones",
            "proto.*",
            "test*",
            "trusted.conf",
            "ns2/K*",
            "ns2/bl.tld2.db",
            "ns2/tld2s.db",
            "ns3/bl-2.db",
            "ns3/bl-cname.db",
            "ns3/bl-disabled.db",
            "ns3/bl-drop.db",
            "ns3/bl-garden.db",
            "ns3/bl-given.db",
            "ns3/bl-no-op.db",
            "ns3/bl-nodata.db",
            "ns3/bl-nxdomain.db",
            "ns3/bl-passthru.db",
            "ns3/bl-tcp-only.db",
            "ns3/bl-wildcname.db",
            "ns3/bl.db",
            "ns3/bl.tld2.db",
            "ns3/evil-cname.db",
            "ns3/fast-expire.db",
            "ns3/manual-update-rpz.db",
            "ns3/mixed-case-rpz.db",
            "ns3/named.conf.tmp",
            "ns3/named.stats",
            "ns3/slow-rpz.db",
            "ns3/wild-cname.db",
            "ns5/bl.db",
            "ns5/empty.db",
            "ns5/empty.db.jnl",
            "ns5/example.db",
            "ns5/expire.conf",
            "ns5/fast-expire.db",
            "ns5/named.stats",
            "ns5/policy2.db",
            "ns5/policy2.db.jnl",
            "ns5/rpz-switch",
            "ns6/bl.tld2s.db",
            "ns6/empty.db",
            "ns6/named.stats",
            "ns7/policy2.db",
            "ns8/manual-update-rpz.db",
        ]
    ),
]


def test_rpz_dnsrps(run_tests_sh):
    with open("dnsrps.conf", "w", encoding="utf-8") as conf:
        conf.writelines(
            [
                "dnsrps-options { log-level 3 };"
                "dnsrps-enable yes;"
                'dnsrps-library "../../rpz/testlib/.libs/libdummyrpz.so";'
            ]
        )
    run_tests_sh()



# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

import os
import time
from typing import Any, Callable, Optional

import dns.query
import dns.message

import isctest.log
from isctest.compat import dns_rcode

QUERY_TIMEOUT = 10


def generic_query(
    query_func: Callable[..., Any],
    message: dns.message.Message,
    ip: str,
    port: Optional[int] = None,
    source: Optional[str] = None,
    timeout: int = QUERY_TIMEOUT,
    attempts: int = 10,
    expected_rcode: dns_rcode = None,
    verify: bool = False,
    log_query: bool = True,
    log_response: bool = True,
) -> Any:
    if port is None:
        if query_func.__name__ == "tls":
            port = int(os.environ["TLSPORT"])
        else:
            port = int(os.environ["PORT"])

    query_args = {
        "q": message,
        "where": ip,
        "timeout": timeout,
        "port": port,
        "source": source,
    }
    if query_func.__name__ == "tls":
        query_args["verify"] = verify

    res = None
    for attempt in range(attempts):
        log_msg = (
            f"isc.query.{query_func.__name__}(): ip={ip}, port={port}, source={source}, "
            f"timeout={timeout}, attempts left={attempts-attempt}"
        )
        if log_query:
            log_msg += f"\n{message.to_text()}"
            log_query = False  # only log query on first attempt
        isctest.log.debug(log_msg)
        try:
            res = query_func(**query_args)
        except (dns.exception.Timeout, ConnectionRefusedError) as e:
            isctest.log.debug(
                f"isc.query.{query_func.__name__}(): the '{e}' exception raised"
            )
        else:
            if log_response:
                isctest.log.debug(
                    f"isc.query.{query_func.__name__}(): response\n{res.to_text()}"
                )
            if res.rcode() == expected_rcode or expected_rcode is None:
                return res
        time.sleep(1)

    if expected_rcode is not None:
        last_rcode = dns_rcode.to_text(res.rcode()) if res else None
        isctest.log.debug(
            f"isc.query.{query_func.__name__}(): expected rcode={dns_rcode.to_text(expected_rcode)}, last rcode={last_rcode}"
        )
    raise dns.exception.Timeout


def udp(*args, **kwargs) -> Any:
    return generic_query(dns.query.udp, *args, **kwargs)


def tcp(*args, **kwargs) -> Any:
    return generic_query(dns.query.tcp, *args, **kwargs)


def tls(*args, **kwargs) -> Any:
    try:
        return generic_query(dns.query.tls, *args, **kwargs)
    except TypeError as e:
        raise RuntimeError(
            "dnspython 2.5.0 or newer is required for isctest.query.tls()"
        ) from e


def create(
    qname,
    qtype,
    qclass=dns.rdataclass.IN,
    dnssec: bool = True,
    cd: bool = False,
    ad: bool = True,
) -> dns.message.Message:
    """Create DNS query with defaults suitable for our tests."""
    msg = dns.message.make_query(
        qname, qtype, qclass, use_edns=True, want_dnssec=dnssec
    )
    msg.flags = dns.flags.RD
    if ad:
        msg.flags |= dns.flags.AD
    if cd:
        msg.flags |= dns.flags.CD
    return msg



# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

from re import compile as Re

import isctest


def check_soa(ns, serial):
    msg = isctest.query.create("test.", "SOA")
    res = isctest.query.udp(msg, ns.ip)
    isctest.check.noerror(res)
    assert len(res.answer) == 1
    assert (
        res.answer[0].to_text()
        == f"test. 60 IN SOA ns.test. op.test. {serial} 100 100 300 60"
    )


def wait_for_initial_xfrin(ns):
    with ns.watch_log_from_start() as watcher:
        watcher.wait_for_line("Transfer status: success")
    check_soa(ns, 1)


def wait_for_sending_notify(ns1, ns, key_name):
    pattern = Re(
        f"zone test/IN: sending notify to {ns.ip}#[0-9]+ : TSIG \\({key_name}\\)"
    )
    with ns1.watch_log_from_start() as watcher:
        watcher.wait_for_line(pattern)


def test_xfer_servers_list(ns1, ns2, ns3, ns4, templates):
    # First, wait for ns2, ns3 and ns4 to xfrin foo.fr and answer it
    wait_for_initial_xfrin(ns2)
    wait_for_initial_xfrin(ns3)
    wait_for_initial_xfrin(ns4)

    # ns1 initially notifies the secondaries using the respectively configured keys
    # - 10.53.0.2 has the key defined where `secondaries` is used
    # - 10.53.0.3 has the key directly after its IP address
    # - 10.53.0.4 has the key defined where `secondariesbis` is used
    #   (inside `secondaries`), so it uses this one instead of the one
    #   defined where `secondaries` is used.
    # Because the order notification are sent doesn't matter here, we can't use wait_for_sequence
    seq = [(ns2, "notifykey2"), (ns3, "notifykey3"), (ns4, "notifykey4")]
    for ns, key_name in seq:
        wait_for_sending_notify(ns1, ns, key_name)

    # Then, ns1 update foo.fr. It notifies ns2, ns3 and ns4 about it
    templates.render("ns1/test.db", {"serial": 2})
    with ns2.watch_log_from_here() as ns2_watcher, ns3.watch_log_from_here() as ns3_watcher, ns4.watch_log_from_here() as ns4_watcher:
        ns1.rndc("reload")
        ns2_watcher.wait_for_line("Transfer status: success")
        ns3_watcher.wait_for_line("Transfer status: success")
        ns4_watcher.wait_for_line("Transfer status: success")
    check_soa(ns2, 2)
    check_soa(ns3, 2)
    check_soa(ns4, 2)



#!/usr/bin/python3

# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

import pytest
import isctest

# isctest.asyncserver requires dnspython >= 2.0.0
pytest.importorskip("dns", minversion="2.0.0")

import dns.message

pytestmark = pytest.mark.extra_artifacts(
    [
        "ans*/ans.run",
    ]
)


def test_connreset():
    msg = dns.message.make_query(
        "sub.example.", "A", want_dnssec=True, use_edns=0, payload=1232
    )
    res = isctest.query.udp(msg, "10.53.0.2")
    isctest.check.servfail(res)



#!/usr/bin/python
# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

# catzhash.py: generate the SHA-1 hash of a domain name in wire format.
#
# This can be used to determine the label to use in a catalog zone to
# represent the specified zone. For example, the zone
# "domain.example" can be represented in a catalog zone called
# "catalog.example" by adding the following record:
#
# 5960775ba382e7a4e09263fc06e7c00569b6a05c.zones.catalog.example. \
#    IN PTR domain.example.
#
# The label "5960775ba382e7a4e09263fc06e7c00569b6a05c" is the output of
# this script when run with the argument "domain.example".

import sys
import hashlib
import dns.name

if len(sys.argv) < 2:
    print("Usage: %s name" % sys.argv[0])

NAME = dns.name.from_text(sys.argv[1]).to_wire()
print(hashlib.sha1(NAME).hexdigest())



#!/usr/bin/python3
# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.
"""
isctest.name.ZoneAnalyzer self-test
Generate insane test zone and check expected output of ZoneAnalyzer utility class
"""


import collections
import itertools
from pathlib import Path

import dns.name
from dns.name import Name
import pytest

import isctest
import isctest.name

pytest.importorskip("dns", minversion="2.3.0")

# set of properies present in the tested zone - read by tests_zone_analyzer.py
CATEGORIES = frozenset(
    [
        "all_existing_names",
        "delegations",
        "dnames",
        "ents",
        "occluded",
        "reachable",
        "reachable_delegations",
        "reachable_dnames",
        "reachable_wildcards",
        "reachable_wildcard_parents",
        "wildcards",
    ]
)


pytestmark = pytest.mark.extra_artifacts(["analyzer.db"])
SUFFIX = dns.name.from_text("nsec3.example.")

LABELS = (b"*", b"dname", b"ent", b"ns", b"txt")
LABEL2RRTYPE = {  # leftmost label encodes RR type we will synthesize for given name
    b"*": "TXT",
    b"dname": "DNAME",
    b"ent": None,  # ENT is not really a 'type'
    b"ns": "NS",
    b"txt": "TXT",
}
LABEL2TAGS = {  # leftmost label encodes 'initial' meaning of a complete name
    b"*": {"wildcards"},
    b"dname": {"dnames"},
    b"ns": {"delegations"},
    b"txt": set(),  # perhaps reachable, perhaps not, we need to decide based on other labels
}


def name2tags(name):
    """
    Decode meaning hidden in labels and their relationships
    and return all tags expected from ZoneAnalyzer
    """
    tags = LABEL2TAGS[name[0]].copy()

    parent_labels = name[1:]
    if b"ns" in parent_labels or b"dname" in parent_labels:
        tags.add("occluded")

    if "occluded" not in tags:
        tags.add("all_existing_names")
        if "delegations" in tags:
            # delegations are ambiguous and don't count as 'reachable'
            tags.add("reachable_delegations")
        elif "dnames" in tags:
            tags.add("reachable")
            tags.add("reachable_dnames")
        elif "wildcards" in tags:
            tags.add("reachable")
            tags.add("reachable_wildcards")
        else:
            tags.add("reachable")

    return tags


def gen_node(nodes, labels):
    name = Name(labels)
    nodes[name] = name2tags(name)


def add_ents(nodes):
    """
    Non-occluded nodes with 'ent' as a parent label imply existence of 'ent' nodes.
    """
    new_ents = {}
    for name, tags in nodes.items():
        if "occluded" in tags:
            continue

        # check if any parent is ENT
        entidx = 1
        while True:
            try:
                entidx = name.labels.index(b"ent", entidx)
            except ValueError:
                break
            entname = Name(name[entidx:])
            new_ents[entname] = {"all_existing_names", "ents"}
            entidx += 1

    return new_ents


def tag_wildcard_parents(nodes):
    """
    Non-occluded nodes with '*' as a leftmost label tag their immediate parent
    nodes as 'reachable_wildcard_parents'.
    """
    for name, tags in nodes.items():
        if "occluded" in tags or not name.is_wild():
            continue

        parent_name = Name(name[1:])
        nodes[parent_name].add("reachable_wildcard_parents")


def is_non_ent(labels):
    """
    Filter out nodes with 'ent' at leftmost position. To become ENT a name must
    not have data by itself but have some other node defined underneath it,
    and must not be occluded, which is something itertools.product() cannot
    decide.
    """
    return labels[0] != b"ent"


def gen_zone(nodes):
    """
    Generate zone file in text format.

    All names are relative.
    Right-hand side of RRs contains dot-separated list of categories a node
    belongs to (except for zone origin).
    """
    for name, tags in sorted(nodes.items()):
        if len(name) == 0:
            # origin, very special case
            yield "@\tSOA\treachable. origin-special-case. 0 0 0 0 0\n"
            yield "@\tNS\treachable.\n"
            yield "@\tA\t192.0.2.1\n"
            continue

        rrtype = LABEL2RRTYPE[name[0]]
        if rrtype is None:  # ENT
            prefix = "; "
        else:
            prefix = ""
        assert tags
        yield f"{prefix}{name}\t{rrtype}\t{'.'.join(sorted(tags))}.\n"


def gen_expected_output(nodes):
    """
    {category: set(names)} mapping used by the pytest check
    """
    categories = collections.defaultdict(set)
    for name, tags in nodes.items():
        for tag in tags:
            categories[tag].add(name)

    assert set(categories.keys()) == CATEGORIES, (
        "CATEGORIES needs updating",
        CATEGORIES.symmetric_difference(set(categories.keys())),
    )

    return categories


def generate_test_data():
    """
    Prepare the analyzer.db zone file in the current working directory and
    return the expected attribute values for the ZoneAnalyzer instance that
    will be tested using that file.
    """
    nodes = {}

    for length in range(1, len(LABELS) + 1):
        for labelseq in filter(is_non_ent, itertools.product(LABELS, repeat=length)):
            gen_node(nodes, labelseq)

    # special-case to make this look as a valid DNS zone - it needs zone origin node
    nodes[Name([])] = {"all_existing_names", "reachable"}

    nodes.update(add_ents(nodes))
    tag_wildcard_parents(nodes)

    with open("analyzer.db", "w", encoding="ascii") as outf:
        outf.writelines(gen_zone(nodes))

    return gen_expected_output(nodes)


if __name__ == "__main__":
    generate_test_data()


@pytest.fixture(scope="module")
def analyzer_fixture():
    expected_results = generate_test_data()  # creates the "analyzer.db" file
    analyzer = isctest.name.ZoneAnalyzer.read_path(Path("analyzer.db"), origin=SUFFIX)
    return expected_results, analyzer


# pylint: disable=redefined-outer-name
@pytest.mark.parametrize("category", sorted(CATEGORIES))
def test_analyzer_attrs(category, analyzer_fixture):
    expected_results, analyzer = analyzer_fixture
    # relativize results to zone name to make debugging easier
    results = {name.relativize(SUFFIX) for name in getattr(analyzer, category)}
    assert results == expected_results[category]



"""
Copyright (C) Internet Systems Consortium, Inc. ("ISC")

SPDX-License-Identifier: MPL-2.0

This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0.  If a copy of the MPL was not distributed with this
file, you can obtain one at https://mozilla.org/MPL/2.0/.

See the COPYRIGHT file distributed with this work for additional
information regarding copyright ownership.
"""

from typing import AsyncGenerator

import dns.name
import dns.rcode
import dns.rdataclass
import dns.rdatatype
import dns.rrset
import dns.zone

from isctest.asyncserver import (
    AsyncDnsServer,
    DnsResponseSend,
    DomainHandler,
    QueryContext,
    ResponseAction,
)

try:
    dns_namerelation_equal = dns.name.NameRelation.EQUAL
    dns_namerelation_subdomain = dns.name.NameRelation.SUBDOMAIN
except AttributeError:  # dnspython < 2.3.0 compat
    dns_namerelation_equal = dns.name.NAMERELN_EQUAL  # type: ignore
    dns_namerelation_subdomain = dns.name.NAMERELN_SUBDOMAIN  # type: ignore


def get_dname_rrset_at_name(
    zone: dns.zone.Zone, name: dns.name.Name
) -> dns.rrset.RRset:
    node = zone.get_node(name)
    assert node
    dname = node.get_rdataset(dns.rdataclass.IN, dns.rdatatype.DNAME)
    assert dname
    rrset = dns.rrset.RRset(name, dname.rdclass, dname.rdtype)
    rrset.update(dname)
    return rrset


class CnameThenDnameHandler(DomainHandler):
    """
    For certain trigger QNAMEs, insert a DNAME RRset after the CNAME chain
    prepared from zone data.
    """

    domains = ["example.broken."]

    async def get_responses(
        self, qctx: QueryContext
    ) -> AsyncGenerator[ResponseAction, None]:
        assert qctx.zone
        assert qctx.zone.origin

        relative_qname = qctx.qname.relativize(qctx.zone.origin)
        if relative_qname.labels[-1].endswith(b"-then-dname"):
            last_cname = qctx.response.answer[-1]
            assert last_cname.rdtype == dns.rdatatype.CNAME
            dname_owner = last_cname.name.parent()
            dname_rrset = get_dname_rrset_at_name(qctx.zone, dname_owner)
            qctx.response.answer.append(dname_rrset)

        yield DnsResponseSend(qctx.response)


class Cve202125215(DomainHandler):
    """
    Attempt to trigger the resolver variant of CVE-2021-25215.  A `named`
    instance cannot be used for serving the DNAME records returned by this
    response handler as a version of BIND 9 vulnerable to CVE-2021-25215 would
    crash while answering the queries sent by the tested resolver.
    """

    domains = ["example.dname."]

    async def get_responses(
        self, qctx: QueryContext
    ) -> AsyncGenerator[ResponseAction, None]:
        assert qctx.zone
        assert qctx.zone.origin

        self_example_dname = dns.name.Name(["self"]).concatenate(qctx.zone.origin)
        dname_rrset = get_dname_rrset_at_name(qctx.zone, self_example_dname)

        relation, _, _ = qctx.qname.fullcompare(self_example_dname)

        if relation in (dns_namerelation_equal, dns_namerelation_subdomain):
            del qctx.response.authority[:]
            qctx.response.set_rcode(dns.rcode.NOERROR)

        if relation == dns_namerelation_subdomain:
            qctx.response.answer.append(dname_rrset)
            cname_rrset = dns.rrset.from_text(
                qctx.qname,
                60,
                qctx.qclass,
                dns.rdatatype.CNAME,
                self_example_dname.to_text(),
            )
            qctx.response.answer.append(cname_rrset)

        yield DnsResponseSend(qctx.response)


def main() -> None:
    server = AsyncDnsServer(acknowledge_manual_dname_handling=True, default_aa=True)
    server.install_response_handlers([CnameThenDnameHandler(), Cve202125215()])
    server.run()


if __name__ == "__main__":
    main()



# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

import pytest


@pytest.mark.requires_zones_loaded("ns1")
def test_spf_log(ns1):
    for msg in (
        "zone spf/IN: 'y.spf' found type SPF record but no SPF TXT record found",
        "zone warn/IN: 'y.warn' found type SPF record but no SPF TXT record found",
        "zone spf/IN: loaded serial 0",
        "zone warn/IN: loaded serial 0",
        "zone nowarn/IN: loaded serial 0",
    ):
        assert msg in ns1.log

    for msg in (
        "zone nowarn/IN: 'y.nowarn' found type SPF record but no SPF TXT record found",
        "zone spf/IN: 'spf' found type SPF record but no SPF TXT record found",
        "zone warn/IN: 'warn' found type SPF record but no SPF TXT record found",
        "zone nowarn/IN: 'nowarn' found type SPF record but no SPF TXT record found",
    ):
        assert msg not in ns1.log



# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

import pytest

pytestmark = [
    pytest.mark.algorithm_set("ecc_default"),
    pytest.mark.extra_artifacts(
        [
            "delv.*",
            "dig.out.*",
            "dsset-*",
            "rndc.out.*",
            "signer.out.*",
            "ns*/dsset-*",
            "ns*/*.bk",
            "ns*/*.jnl",
            "ns*/K*",
            "ns*/island.conf",
            "ns*/managed.conf",
            "ns*/named.secroots",
            "ns*/private.conf",
            "ns*/signer.out.*",
            "ns*/trusted.conf",
            "ns*/*.signed",
            "ns1/managed.key",
            "ns1/managed.key.id",
            "ns1/root.db.orig",
            "ns1/root.db.tmp",
            "ns1/zone.key",
            "ns2/managed-keys.bind",
            "ns2/managed1.conf",
            "ns3/broken.conf",
            "ns3/managed-keys.bind",
            "ns4/nope",
            "ns5/managed-keys.bind",
            "ns5/named.args",
            "ns7/view1.mkeys",
            "ns7/view2.mkeys",
        ]
    ),
]


@pytest.mark.flaky(max_runs=2)  # GL#3098
def test_mkeys(run_tests_sh):
    run_tests_sh()



# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

import os

from .all import ALL
from .algorithms import init_crypto_supported, set_algorithm_set
from .features import init_features
from .openssl import parse_openssl_config
from .. import log


def init_vars():
    """Initializes the environment variables."""
    init_features()
    init_crypto_supported()
    set_algorithm_set(os.getenv("ALGORITHM_SET"))
    parse_openssl_config(ALL["OPENSSL_CONF"])

    os.environ.update(ALL)
    log.debug("setting following env vars: %s", ", ".join([str(key) for key in ALL]))



"""
Copyright (C) Internet Systems Consortium, Inc. ("ISC")

SPDX-License-Identifier: MPL-2.0

This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0.  If a copy of the MPL was not distributed with this
file, you can obtain one at https://mozilla.org/MPL/2.0/.

See the COPYRIGHT file distributed with this work for additional
information regarding copyright ownership.
"""

from typing import AsyncGenerator

import dns.rcode
import dns.rdatatype

from isctest.asyncserver import (
    AsyncDnsServer,
    DnsResponseSend,
    DomainHandler,
    QueryContext,
    ResponseAction,
)

from qmin_ans import DelayedResponseHandler, EntRcodeChanger, QueryLogHandler, log_query


class QueryLogger(QueryLogHandler):
    domains = [
        "1.1.1.1.8.2.6.0.1.0.0.2.ip6.arpa.",
        "icky.ptang.zoop.boing.good.",
    ]


class StaleHandler(DomainHandler):
    """
    The test code relies on this server returning non-minimal (i.e. including
    address records in the ADDITIONAL section) responses to NS queries for
    `b.stale` and `a.b.stale`.  While this logic (returning non-minimal
    responses to NS queries) could be implemented in AsyncDnsServer itself,
    doing so breaks a lot of other checks in this system test.  Therefore, only
    these two zones behave in this particular way, thanks to a custom response
    handler implemented below.
    """

    domains = ["b.stale", "a.b.stale"]

    async def get_responses(
        self, qctx: QueryContext
    ) -> AsyncGenerator[ResponseAction, None]:
        log_query(qctx)

        if qctx.qtype == dns.rdatatype.NS:
            assert qctx.zone
            assert qctx.response.answer[0]

            for nameserver in qctx.response.answer[0]:
                if not nameserver.target.is_subdomain(qctx.response.answer[0].name):
                    continue
                glue_a = qctx.zone.get_rrset(nameserver.target, dns.rdatatype.A)
                if glue_a:
                    qctx.response.additional.append(glue_a)
                glue_aaaa = qctx.zone.get_rrset(nameserver.target, dns.rdatatype.AAAA)
                if glue_aaaa:
                    qctx.response.additional.append(glue_aaaa)

        yield DnsResponseSend(qctx.response)


class IckyPtangZoopBoingBadHandler(EntRcodeChanger):
    domains = ["icky.ptang.zoop.boing.bad."]
    rcode = dns.rcode.NXDOMAIN


class IckyPtangZoopBoingUglyHandler(EntRcodeChanger):
    domains = ["icky.ptang.zoop.boing.ugly."]
    rcode = dns.rcode.FORMERR


class IckyPtangZoopBoingSlowHandler(DelayedResponseHandler):
    domains = ["icky.ptang.zoop.boing.slow."]
    delay = 0.4


def main() -> None:
    server = AsyncDnsServer()
    server.install_response_handlers(
        [
            QueryLogger(),
            StaleHandler(),
            IckyPtangZoopBoingBadHandler(),
            IckyPtangZoopBoingUglyHandler(),
            IckyPtangZoopBoingSlowHandler(),
        ]
    )
    server.run()


if __name__ == "__main__":
    main()



#!/usr/bin/python3

# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

# pylint: disable=unused-variable

import socket
import time

import pytest

# isctest.asyncserver requires dnspython >= 2.0.0
pytest.importorskip("dns", minversion="2.0.0")

import dns.message
import dns.query
import dns.tsigkeyring

pytestmark = pytest.mark.extra_artifacts(
    [
        "ans*/ans.run",
        "ns1/named-fips.conf",
    ]
)

TIMEOUT = 10


def create_msg(qname, qtype, edns=-1):
    msg = dns.message.make_query(qname, qtype, use_edns=edns)
    return msg


def timeout():
    return time.time() + TIMEOUT


def create_socket(host, port):
    sock = socket.create_connection((host, port), timeout=10)
    sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, True)
    return sock


def test_tsig_badtime(named_port):
    with create_socket("10.53.0.1", named_port) as sock:
        msg = create_msg("a.example.", "A")

        keyring = dns.tsigkeyring.from_text(
            {
                "sha256": "R16NojROxtxH/xbDl//ehDsHm5DjWTQ2YXV+hGC2iBY=",
            }
        )

        msg.use_tsig(keyring, keyname="sha256", fudge=0)

        wire = msg.to_wire()
        assert len(wire) > 0

        time.sleep(3)

        dns.query.send_tcp(sock, wire, timeout())
        with pytest.raises(dns.tsig.PeerBadTime):
            dns.query.receive_tcp(sock, timeout(), keyring=keyring)



# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

import os

# pylint: disable=import-error
from .autoconf import AC_VARS  # type: ignore

# pylint: enable=import-error


SYSTEM_TEST_DIR_GIT_PATH = "bin/tests/system"

DIR_VARS = {
    "builddir": f"{AC_VARS['TOP_BUILDDIR']}/{SYSTEM_TEST_DIR_GIT_PATH}",
    "srcdir": f"{AC_VARS['TOP_SRCDIR']}/{SYSTEM_TEST_DIR_GIT_PATH}",
    "HYPOTHESIS_STORAGE_DIRECTORY": f"{AC_VARS['TOP_BUILDDIR']}/{SYSTEM_TEST_DIR_GIT_PATH}/.hypothesis",
    "SYSTESTDIR": None,
}


def set_system_test_name(name: str):
    DIR_VARS["SYSTESTDIR"] = name
    os.environ["SYSTESTDIR"] = name



# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

from . import check
from . import instance
from . import query
from . import kasp
from . import run
from . import template
from . import log
from . import vars  # pylint: disable=redefined-builtin

# isctest.hypothesis is intentionally NOT imported, because it detects proper
# hypothesis support and instructs pytest to skip the tests otherwise. It
# should be manually imported only in the modules that require hypothesis.

# isctest.mark module is intentionally NOT imported, because it relies on
# environment variables which might not be set at the time of import of the
# `isctest` package. To use the marks, manual `import isctest.mark` is needed
# instead.
